import asyncio
from glob import glob
import os
import time
from typing import Callable, List, Optional
import zipfile
from learning_loop_node.loop import loop
from urllib.parse import urljoin
from requests import Session
import aiohttp
import logging
from icecream import ic
import shutil


class LiveServerSession(Session):
    """https://stackoverflow.com/a/51026159/364388"""

    def __init__(self, *args, **kwargs):
        super(LiveServerSession, self).__init__(*args, **kwargs)
        self.prefix_url = loop.base_url

    def request(self, method, url, *args, **kwargs):
        url = urljoin(self.prefix_url, url)
        headers = {}
        if 'token' in url:
            return super(LiveServerSession, self).request(method, url, *args, **kwargs)

        headers = loop.get_headers()
        return super(LiveServerSession, self).request(method, url, headers=headers, *args, **kwargs)


def get_files_in_folder(folder: str):
    files = [entry for entry in glob(f'{folder}/**/*', recursive=True) if os.path.isfile(entry)]
    files.sort()
    return files


async def get_latest_model_id() -> str:
    async with loop.get(f'api/zauberzeug/projects/pytest/trainings') as response:
        assert response.status == 200
        trainings = await response.json()
        return trainings['charts'][0]['data'][0]['model_id']


async def assert_upload_model_with_id(file_paths: Optional[List[str]] = None, format: str = 'mocked', model_id: Optional[str] = None) -> str:
    data = prepare_formdata(file_paths)

    async with loop.put(f'api/zauberzeug/projects/pytest/models/{model_id}/{format}/file', data) as response:
        if response.status != 200:
            msg = f'unexpected status code {response.status} while putting model'
            logging.error(msg)
            raise (Exception(msg))
        model = await response.json()

        return model['id']


def prepare_formdata(file_paths: Optional[List[str]]) -> aiohttp.FormData:
    module_path = os.path.dirname(os.path.realpath(__file__))
    if not file_paths:
        file_paths = [f'{module_path}/test_data/file_1.txt',
                      f'{module_path}/test_data/file_2.txt',
                      f'{module_path}/test_data/model.json']

    data = [('files', open(path, 'rb')) for path in file_paths]
    data = aiohttp.FormData()
    for path in file_paths:
        data.add_field('files',  open(path, 'rb'))
    return data


def unzip(file_path, target_folder):
    shutil.rmtree(target_folder, ignore_errors=True)
    os.makedirs(target_folder)
    with zipfile.ZipFile(file_path, 'r') as zip:
        zip.extractall(target_folder)


async def condition(condition: Callable, *, timeout: float = 1.0, interval: float = 0.1):
    start = time.time()
    while not condition():
        if time.time() > start + timeout:
            raise TimeoutError(f'condition {condition} took longer than {timeout}s')
        await asyncio.sleep(interval)


def update_attributes(obj, **kwargs) -> None:
    if isinstance(obj, dict):
        _update_attribute_dict(obj, **kwargs)
    else:
        _update_attribute_class_instance(obj, **kwargs)


def _update_attribute_class_instance(obj, **kwargs) -> None:
    for key, value in kwargs.items():
        if hasattr(obj, key):
            setattr(obj, key, value)
        else:
            raise ValueError(f"Object of type '{type(obj)}' does not have a property '{key}'.")


def _update_attribute_dict(obj: dict, **kwargs) -> None:
    for key, value in kwargs.items():
        obj[key] = value



from typing import Dict, List, Optional, Union
from learning_loop_node.model_information import ModelInformation
from learning_loop_node.trainer import Trainer
from learning_loop_node.trainer.model import BasicModel, PretrainedModel
import subprocess
import asyncio
import logging
import time
from learning_loop_node.context import Context
import socketio


class TestingTrainer(Trainer):
    __test__ = False

    def __init__(self, can_resume=False) -> None:
        super().__init__('mocked')
        self._can_resume = can_resume
        self.has_new_model = False

    async def start_training(self) -> None:
        self.executor.start('while true; do sleep 1; done')

    @property
    def provided_pretrained_models(self) -> List[PretrainedModel]:
        return [
            PretrainedModel(name='small', label='Small', description='a small model'),
            PretrainedModel(name='medium', label='Medium', description='a medium model'),
            PretrainedModel(name='large', label='Large', description='a large model')]

    async def start_training(self, model: str = 'model.model') -> None:
        self.executor.start('while true; do sleep 1; done')

    async def start_training_from_scratch(self, id: str) -> None:
        await self.start_training(model=f'model_{id}.pt')

    def get_new_model(self) -> Optional[BasicModel]:
        if self.has_new_model:
            return BasicModel(confusion_matrix={})

        return None

    def on_model_published(self, basic_model: BasicModel) -> None:
        pass

    async def _prepare(self) -> None:
        await super()._prepare()
        await asyncio.sleep(0.1)  # give tests a bit time to to check for the state

    async def _download_model(self) -> None:
        await super()._download_model()
        await asyncio.sleep(0.1)  # give tests a bit time to to check for the state

    async def ensure_confusion_matrix_synced(self, trainer_node_uuid: str, sio_client: socketio.AsyncClient):
        await asyncio.sleep(0.1)  # give tests a bit time to to check for the state
        await super().ensure_confusion_matrix_synced(trainer_node_uuid, sio_client)
        await asyncio.sleep(0.1)  # give tests a bit time to to check for the state

    async def upload_model(self) -> None:
        await asyncio.sleep(0.1)  # give tests a bit time to to check for the state
        result = await super().upload_model()
        await asyncio.sleep(0.1)  # give tests a bit time to to check for the state
        return result

    async def _upload_model(self, context: Context) -> dict:
        await asyncio.sleep(0.1)  # give tests a bit time to to check for the state
        result = await super()._upload_model(context)
        await asyncio.sleep(0.1)  # give tests a bit time to to check for the state
        return result

    def get_latest_model_files(self) -> Union[List[str], Dict[str, List[str]]]:
        time.sleep(1)  # NOTE reduce flakyness in Backend tests du to wrong order of events.
        fake_weight_file = '/tmp/weightfile.weights'
        with open(fake_weight_file, 'wb') as f:
            f.write(b'\x42')

        more_data_file = '/tmp/some_more_data.txt'
        with open(more_data_file, 'w') as f:
            f.write('zweiundvierzig')
        return {'mocked': [fake_weight_file, more_data_file], 'mocked_2': [fake_weight_file, more_data_file]}

    def can_resume(self) -> bool:
        return self._can_resume

    async def resume(self) -> None:
        return await self.start_training()

    async def _detect(self, model_information: ModelInformation, images:  List[str], model_folder: str) -> List:
        detections = []
        return detections

    async def clear_training_data(self, training_folder: str) -> None:
        return



from .detections import Detections
from .box_detection import BoxDetection
from .point_detection import PointDetection
from .segmentation_detection import SegmentationDetection, Shape, Point
from .outbox.outbox import Outbox



import sys
if sys.version_info.major >= 3 and sys.version_info.minor >= 7:  # most code needs at least python 3.7
    from .trainer.trainer_node import TrainerNode
    from .converter.converter_node import ConverterNode

from .context import Context
from .detector.detector_node import DetectorNode
from .detector.detector import Detector
from .model_information import ModelInformation
from .data_classes.category import CategoryType
from .globals import GLOBALS
from .loop import loop

from . import log_conf

log_conf.init()




from fastapi.encoders import jsonable_encoder
from learning_loop_node.node import Node
from learning_loop_node.status import State, AnnotationNodeStatus
from learning_loop_node.context import Context
from learning_loop_node.annotation_node.annotation_tool import AnnotationTool
from learning_loop_node.annotation_node.data_classes import EventType, UserInput
from learning_loop_node.rest.downloader import DataDownloader
from learning_loop_node.rest.downloader import node_helper
import logging
from icecream import ic
from ..socket_response import SocketResponse


class AnnotationNode(Node):
    tool: AnnotationTool
    histories: dict = {}

    def __init__(self, name: str, uuid: str, tool: AnnotationTool):
        super().__init__(name, uuid)
        self.tool = tool

        @self.sio_client.on('handle_user_input')
        async def on_handle_user_input(user_input):
            return await self.handle_user_input(user_input)

        @self.sio_client.on('user_logout')
        async def on_logout_user(sid):
            self.reset_history(sid)
            return self.tool.logout_user(sid)

    async def handle_user_input(self, user_input) -> str:
        input = UserInput.parse_obj(user_input)

        if input.data.key_up == 'Escape':
            self.reset_history(input.frontend_id)
            return ''

        await self.download_image(input.data.context, input.data.image_uuid)
        history = self.get_history(input.frontend_id)
        try:
            tool_result = await self.tool.handle_user_input(input, history)
        except:
            self.reset_history(input.frontend_id)
            raise

        if tool_result.annotation:
            result = await self.sio_client.call('update_segmentation_annotation', (input.data.context.organization,
                                                                                   input.data.context.project, jsonable_encoder(tool_result.annotation)), timeout=2)

        return tool_result.svg

    def reset_history(self, frontend_id: str) -> None:
        try:
            del self.histories[frontend_id]
        except:
            pass

    def get_history(self, frontend_id: str) -> dict:
        if not frontend_id in self.histories.keys():
            self.histories[frontend_id] = self.tool.create_empty_history()

        return self.histories[frontend_id]

    async def send_status(self):
        status = AnnotationNodeStatus(
            id=self.uuid,
            name=self.name,
            state=State.Online,
            capabilities=['segmentation']
        )

        logging.info(f'sending status {status}')
        result = await self.sio_client.call('update_annotation_node', jsonable_encoder(status), timeout=2)
        response = SocketResponse.from_dict(result)

        if not response.success:
            logging.error(f'Error for updating: Response from loop was : {response.__dict__}')

    async def download_image(self, context: Context, uuid: str):
        project_folder = Node.create_project_folder(context)
        images_folder = node_helper.create_image_folder(project_folder)

        downloader = DataDownloader(context=context)
        await downloader.download_images([uuid], images_folder)



from typing import List, Optional
from dataclasses import dataclass, field
from ..status import State
from .model import PretrainedModel


@dataclass
class TrainingStatus():
    id: str
    name: str
    state: Optional[State]
    errors: Optional[dict]
    uptime: Optional[int]
    progress: Optional[float]

    train_image_count: Optional[int] = None
    test_image_count: Optional[int] = None
    skipped_image_count: Optional[int] = None
    pretrained_models: List[PretrainedModel] = field(default_factory=list)
    hyperparameters: Optional[str] = None
    architecture: Optional[str] = None



from ast import List
from learning_loop_node.trainer.training import Training
import os
import json


def file_path(training: Training) -> str:
    return f'{training.training_folder}/detections.json'


def save(training: Training, detections: List) -> None:
    with open(file_path(training), 'w') as f:
        json.dump(detections, f)


def load(training: Training,) -> List:
    with open(file_path(training), 'r') as f:
        return json.load(f)


def delete(training: Training) -> None:
    if exists(training):
        os.remove(file_path(training))


def exists(training: Training) -> bool:
    return os.path.exists(file_path(training))



import datetime
import logging.config
LOGGING_CONF = {
    'version': 1,
    'disable_existing_loggers': True,  # to make sure this config is used
    'formatters': {
        'default': {
            'format': '%(asctime)s,%(msecs)03d %(levelname)-8s [%(filename)s:%(lineno)d] %(message)s',
            'datefmt': '%Y-%m-%d %H:%M:%S',
        },
    },
    'handlers': {
        'console': {
            'class': 'logging.StreamHandler',
            'formatter': 'default',
            'level': 'DEBUG',
            'stream': 'ext://sys.stdout'
        },
    },
    'loggers': {
        '': {  # this root logger is used for everything without a specific logger
            'handlers': ['console'],
            'level': 'INFO',
            'propagate': False,
        },
    },
}


def init():
    logging.config.dictConfig(LOGGING_CONF)
    logger = logging.getLogger()



from learning_loop_node.detector.box_detection import BoxDetection
from learning_loop_node.trainer.tests.testing_trainer import TestingTrainer
from learning_loop_node.trainer.tests.states import state_helper
from learning_loop_node.trainer.tests.states.state_helper import assert_training_state
from learning_loop_node.trainer import active_training
from learning_loop_node.trainer.trainer import Trainer
from learning_loop_node.trainer.training import Training
from learning_loop_node.context import Context
import asyncio
from learning_loop_node import loop
from fastapi.encoders import jsonable_encoder
error_key = 'upload_detections'


def trainer_has_error(trainer: Trainer):
    return trainer.errors.has_error_for(error_key)


async def create_valid_detection_file(training: Training):
    async with loop.get(f'/api/zauberzeug/projects/demo/data') as response:
        assert response.status == 200, response
        content = await response.json()

    category = content['categories'][0]
    image_id = content['image_ids'][0]
    model_version = '1.2'
    box_detections = [BoxDetection(category['name'], x=1, y=2, width=30, height=40,
                                   net=model_version, confidence=.99, category_id=category['id'])]

    image_entry = {'image_id': image_id, 'box_detections': box_detections,
                   'point_detections': [], 'segmentation_detections': []}
    active_training.detections.save(training, jsonable_encoder([image_entry]))


async def test_upload_successful():
    state_helper.create_active_training_file(training_state='detected')
    trainer = TestingTrainer()
    trainer.training = active_training.load()  # normally done by node

    await create_valid_detection_file(trainer.training)

    await trainer.upload_detections()

    assert trainer.training.training_state == 'ready_for_cleanup'
    assert active_training.load() == trainer.training


async def test_detection_upload_progress_is_stored():
    state_helper.create_active_training_file(training_state='detected')
    trainer = TestingTrainer()
    trainer.training = active_training.load()  # normally done by node

    await create_valid_detection_file(trainer.training)

    await trainer.upload_detections()
    assert active_training.detections_upload_progress.load(trainer.training) == 1


async def test_bad_status_from_LearningLoop():
    state_helper.create_active_training_file(training_state='detected', context=Context(
        organization='zauberzeug', project='some_bad_project'))
    trainer = TestingTrainer()
    trainer.training = active_training.load()  # normally done by node
    active_training.detections.save(trainer.training, [{'some_bad_data': 'some_bad_data'}])

    train_task = asyncio.get_running_loop().create_task(trainer.train(None, None))
    await assert_training_state(trainer.training, 'detection_uploading', timeout=1, interval=0.001)
    await assert_training_state(trainer.training, 'detected', timeout=1, interval=0.001)

    assert trainer_has_error(trainer)
    assert trainer.training.training_state == 'detected'
    assert active_training.load() == trainer.training


async def test_other_errors():
    # e.g. missing detection file
    state_helper.create_active_training_file(training_state='detected')
    trainer = TestingTrainer()
    trainer.training = active_training.load()  # normally done by node

    train_task = asyncio.get_running_loop().create_task(trainer.train(None, None))
    await assert_training_state(trainer.training, 'detection_uploading', timeout=1, interval=0.001)
    await assert_training_state(trainer.training, 'detected', timeout=1, interval=0.001)

    assert trainer_has_error(trainer)
    assert trainer.training.training_state == 'detected'
    assert active_training.load() == trainer.training


async def test_abort_uploading():
    state_helper.create_active_training_file(training_state='detected')
    trainer = TestingTrainer()
    trainer.training = active_training.load()  # normally done by node
    await create_valid_detection_file(trainer.training)

    train_task = asyncio.get_running_loop().create_task(trainer.train(None, None))

    await assert_training_state(trainer.training, 'detection_uploading', timeout=1, interval=0.001)

    trainer.stop()
    await asyncio.sleep(0.1)

    assert trainer.training is None
    assert active_training.exists() is False



from learning_loop_node.tests import test_helper
from learning_loop_node.gdrive_downloader import g_download
from learning_loop_node.rest import downloads
from learning_loop_node.context import Context
import pytest
import logging
import os
from icecream import ic
from learning_loop_node.globals import GLOBALS


@pytest.mark.skip  # NOTE: Brauchen wir diesen Test noch?
async def test_multiple_get_requests_after_post_request_should_not_causes_timeout_error():
    logging.debug('downloading model from gdrive')

    # folder: https://drive.google.com/drive/u/0/folders/1c1oxQwhj-bjqbYjM0B2q7I9ZZe-s2SWD
    # filename: zuckerruebe_roboter_1.5_yolov5_pytorch.zip
    file_id = '1sZWa053fWT9PodrujDX90psmjhFVLyBV'
    destination = '/tmp/model.zip'
    g_download(file_id, destination)
    test_helper.unzip(destination, '/tmp/some_model')

    data = test_helper.prepare_formdata(['/tmp/some_model/model.pt', '/tmp/some_model/model.json'])
    from learning_loop_node.loop import loop
    async with loop.post(f'api/zauberzeug/projects/pytest/models/yolov5_pytorch', data=data) as response:
        if response.status != 200:
            msg = f'unexpected status code {response.status} while putting model'
            logging.error(msg)
            raise (Exception(msg))
        model = await response.json()

    for i in range(10):
        ic('going to download model from loop')

        import shutil
        shutil.rmtree(GLOBALS.data_folder, ignore_errors=True)
        os.makedirs(GLOBALS.data_folder)
        await downloads.download_model(GLOBALS.data_folder, Context(organization='zauberzeug', project='pytest'), model['id'], 'yolov5_pytorch')



import asyncio
from learning_loop_node.context import Context
import traceback
from fastapi_utils.tasks import repeat_every
from fastapi.encoders import jsonable_encoder
from typing import Union
from uuid import uuid4
from icecream import ic
from learning_loop_node.trainer.training import TrainingOut
from .model import Model
from .trainer import Trainer
from .training_status import TrainingStatus
from learning_loop_node.node import Node, State
import logging
from ..socket_response import SocketResponse
from .rest import controls
from learning_loop_node.trainer import active_training
from learning_loop_node.trainer.training import State as TrainingState
from learning_loop_node.trainer import training_syncronizer


class TrainerNode(Node):
    trainer: Trainer
    model_published: bool = False

    def __init__(self, name: str, trainer: Trainer, uuid: str = None):
        super().__init__(name, uuid)
        self.trainer = trainer
        self.include_router(controls.router, tags=["controls"])
        self.train_loop_busy = False
        active_training.init(self.uuid)

        @self.sio_client.on('begin_training')
        async def on_begin_training(organization: str, project: str, details: dict):
            logging.info('received begin_training from server')
            self.trainer.init(Context(organization=organization, project=project), details)
            self.start_training_task()
            return True

        @self.sio_client.on('stop_training')
        async def stop():
            logging.info(f'### on stop_training received. Current state : {self.status.state}')
            loop = asyncio.get_event_loop()
            try:
                self.stop_training()
            except:
                logging.exception('error in stop_training')
            return True

        @self.on_event("startup")
        @repeat_every(seconds=5, raise_exceptions=True, wait_first=False)
        async def continous_send_status():
            try:
                await self.send_status()
            except:
                logging.exception('could not send status state')

        @self.on_event("shutdown")
        async def shutdown():
            await self.shutdown()

        @self.on_event("startup")
        async def resume_training_if_exists():
            self.start_training_task()

    def stop_training(self, save_and_detect: bool = True) -> Union[bool, str]:
        result = self.trainer.stop()

    async def save_model(self, context: Context):
        self.status.reset_error('save_model')
        uploaded_model = None
        try:
            uploaded_model = await self.trainer.save_model(context)
        except Exception as e:
            logging.exception('could not save model')
            self.status.set_error('save_model', f'Could not save model: {str(e)}')

        await self.send_status()
        return uploaded_model

    def start_training_task(self):
        loop = asyncio.get_event_loop()
        loop.create_task(self.trainer.train(self.uuid, self.sio_client))

    async def send_status(self):
        if not self.trainer.training and active_training.exists():
            logging.warning('Found active training, but no its not loaded yet. Skipping this status update.')
            return

        state_for_learning_loop = TrainerNode.state_for_learning_loop(
            self.trainer.training.training_state) if self.trainer.training else State.Idle
        status = TrainingStatus(
            id=self.uuid,
            name=self.name,
            state=state_for_learning_loop,
            errors={},
            uptime=self.training_uptime,
            progress=self.progress
        )

        status.pretrained_models = self.trainer.provided_pretrained_models
        status.architecture = self.trainer.model_architecture

        if self.trainer.training:
            status.train_image_count = self.trainer.training.data.train_image_count()
            status.test_image_count = self.trainer.training.data.test_image_count()
            status.skipped_image_count = self.trainer.training.data.skipped_image_count
            status.hyperparameters = self.trainer.hyperparameters
            status.errors = self.trainer.errors._errors

        logging.info(f'sending status {status}')
        result = await self.sio_client.call('update_trainer', jsonable_encoder(status), timeout=1)
        response = SocketResponse.from_dict(result)

        if not response.success:
            logging.error(f'Error for updating: Response from loop was : {response.__dict__}')
            logging.exception('update trainer failed')

    async def shutdown(self):
        logging.info('shutdown detected, stopping training')
        self.trainer.shutdown()

    def get_state(self):
        if self.trainer.executor is not None and self.trainer.executor.is_process_running():
            return State.Running
        return State.Idle

    @property
    def progress(self) -> Union[float, None]:
        return self.trainer.progress if hasattr(self.trainer, 'progress') else None

    @property
    def training_uptime(self) -> Union[int, None]:
        import time
        now = time.time()
        return now - self.trainer.start_time if self.trainer.start_time else None

    @staticmethod
    def state_for_learning_loop(trainer_state: TrainingState):
        logging.error(trainer_state)

        if trainer_state == TrainingState.Initialized:
            return 'training is initialized'
        if trainer_state == TrainingState.DataDownloading:
            return 'Downloading data'
        if trainer_state == TrainingState.DataDownloaded:
            return 'Data downloaded'
        if trainer_state == TrainingState.TrainModelDownloading:
            return 'Downloading model'
        if trainer_state == TrainingState.TrainModelDownloaded:
            return 'Model downloaded'
        if trainer_state == TrainingState.TrainingRunning:
            return State.Running
        if trainer_state == TrainingState.TrainingFinished:
            return 'Training finished'
        if trainer_state == TrainingState.Detecting:
            return State.Detecting
        if trainer_state == TrainingState.ConfusionMatrixSyncing:
            return 'Syncing confusion matrix'
        if trainer_state == TrainingState.ConfusionMatrixSynced:
            return 'Confusion matrix synced'
        if trainer_state == TrainingState.TrainModelUploading:
            return 'Uploading trained model'
        if trainer_state == TrainingState.TrainModelUploaded:
            return 'Trained model uploaded'
        if trainer_state == TrainingState.Detecting:
            return 'calculating detections'
        if trainer_state == TrainingState.Detected:
            return 'Detections calculated'
        if trainer_state == TrainingState.DetectionUploading:
            return 'Uploading detections'
        if trainer_state == TrainingState.ReadyForCleanup:
            return 'Cleaning training'
        return 'unknown state'



import pytest
from learning_loop_node import node_helper
from typing import List


@pytest.mark.parametrize("image_ids,expected_urls,expected_ids", [
    (['some_id'], ['api/zauberzeug/projects/pytest/images/some_id/main'], ['some_id']),
    (['some_id_1', 'some_id_2'], ['api/zauberzeug/projects/pytest/images/some_id_1/main',
     'api/zauberzeug/projects/pytest/images/some_id_2/main'], ['some_id_1', 'some_id_2']),
    ([], [], [])
])
def test_resource_path_creation(image_ids: List[str], expected_urls: List[str], expected_ids: List['str']):
    urls, ids = node_helper.create_resource_paths('zauberzeug', 'pytest', image_ids)

    assert urls == expected_urls
    assert ids == expected_ids



from fastapi import APIRouter, Request, File, UploadFile, Header
from typing import Optional
import numpy as np
from fastapi.responses import JSONResponse
from icecream import ic
from ...inbox_filter import relevance_filter

router = APIRouter()


@router.post("/detect")
async def http_detect(
    request: Request,
    file: UploadFile = File(...),
    camera_id: Optional[str] = Header(None),
    mac: Optional[str] = Header(None),
    tags: Optional[str] = Header(None),
    autoupload: Optional[str] = Header(None),
):
    """
    Example Usage

        curl --request POST -F 'file=@test.jpg' localhost:8004/detect

        for i in `seq 1 10`; do time curl --request POST -F 'file=@test.jpg' localhost:8004/detect; done

        You can additionally provide the following camera parameters:
          - `autoupload`: configures auto-submission to the learning loop; `filtered` (default), `all`, `disabled` (example curl parameter `-H 'autoupload: all'`)
          - `camera-id`: a string which groups images for submission together (example curl parameter `-H 'camera-id: front_cam'`)
    """
    try:
        np_image = np.fromfile(file.file, np.uint8)
    except:
        raise Exception(f'Uploaded file {file.filename} is no image file.')
    detections = await request.app.get_detections(
        raw_image=np_image,
        camera_id=camera_id or mac or None,
        tags=tags.split(',') if tags else [],
        autoupload=autoupload,
    )
    return JSONResponse(detections)



from learning_loop_node.trainer.trainer_node import TrainerNode
from learning_loop_node.trainer.tests.testing_trainer import TestingTrainer
from learning_loop_node.trainer.training import Training
from uuid import uuid4
from learning_loop_node.context import Context
from learning_loop_node.trainer import active_training
import os
import pytest
import signal


def create_training() -> Training:
    context = Context(organization='zauberzeug', project='demo')
    training = Training(
        id=str(uuid4()),
        context=context,
        project_folder='',
        images_folder='')
    return training


def test_fixture_trainer_node(test_trainer_node):
    assert isinstance(test_trainer_node, TrainerNode)
    assert isinstance(test_trainer_node.trainer, TestingTrainer)


def test_save_load_training():
    training = create_training()
    training.training_state = 'preparing'
    active_training.save(training)

    training = active_training.load()
    assert training.training_state == 'preparing'



import logging
from learning_loop_node.detector import BoxDetection, PointDetection, SegmentationDetection, Shape, Point, Detections
from learning_loop_node import Detector, ModelInformation


class TestingDetector(Detector):
    __test__ = False

    def __init__(self, segmentation_detections: bool = False) -> None:
        super().__init__('mocked')
        self.segmentation_detections = segmentation_detections

    def init(self,  model_info: ModelInformation,  model_root_path: str):
        self.model_info = model_info

    def evaluate(self, image: bytes) -> Detections:
        if self.segmentation_detections:
            return Detections(
                box_detections=[BoxDetection('some_category_name', 1, 2, 3, 4, 'some_model', .42)],
                point_detections=[PointDetection('some_category_name', 10, 12, 'some_model', .42)],
                segmentation_detections=[SegmentationDetection(
                    'some_category_name', Shape(points=[Point(1, 1)]), 'some_model', .42)]
            )

        return Detections(
            box_detections=[BoxDetection('some_category_name', 1, 2, 3, 4, 'some_model', .42)],
            point_detections=[PointDetection('some_category_name', 10, 12, 'some_model', .42)]
        )





class Errors():
    def __init__(self):
        self._errors: dict = {}

    def set(self, key: str, value: str):
        self._errors[key] = value

    def reset(self, key: str):
        try:
            del self._errors[key]
        except AttributeError:
            pass
        except KeyError:
            pass

    def reset_all(self):
        self._errors = {}

    def has_error_for(self, key: str) -> bool:
        return key in self._errors

    def has_error(self) -> bool:
        return self._errors == {}


class TrainingError(Exception):
    def __init__(self, cause: str, *args: object) -> None:
        super().__init__(*args)
        self.cause = cause



from typing import List
from learning_loop_node.detector.box_detection import BoxDetection
from learning_loop_node.detector.detections import Detections
from learning_loop_node.detector.outbox.outbox import Outbox
from learning_loop_node.detector.point_detection import PointDetection
from learning_loop_node.inbox_filter.relevance_filter import RelevanceFilter
import pytest


high_conf_box_detection = BoxDetection('dirt', 0, 0, 100, 100, 'xyz', .9)
high_conf_point_detection = PointDetection('point', 100, 100, 'xyz', .9)
low_conf_point_detection = PointDetection('point', 100, 100, 'xyz', .3)


@pytest.mark.parametrize("detections,reason",
                         [(Detections(box_detections=[high_conf_box_detection]*40, point_detections=[high_conf_point_detection]*40), ['unexpectedObservationsCount']),
                          (Detections(box_detections=[high_conf_box_detection],
                                      point_detections=[high_conf_point_detection]), []),
                          (Detections(box_detections=[high_conf_box_detection]*40, point_detections=[
                              low_conf_point_detection]*40), ['uncertain', 'unexpectedObservationsCount']),
                          (Detections(box_detections=[high_conf_box_detection], point_detections=[low_conf_point_detection]), ['uncertain'])])
def test_unexpected_observations_count(detections: Detections, reason: List[str]):
    filter = RelevanceFilter(Outbox())
    assert filter.learn(detections, raw_image=b'', camera_id='0:0:0:0', tags=[]) == reason



from typing import List, Optional
from pydantic.main import BaseModel
from enum import Enum


class CategoryType(str, Enum):
    Box = 'box'
    Point = 'point'
    Segmentation = 'segmentation'


class Category(BaseModel):
    id: str
    name: str
    description: Optional[str]
    hotkey: Optional[str]
    color: Optional[str]
    type: CategoryType = CategoryType.Box
    point_size: Optional[int]

    @staticmethod
    def from_list(values: List[dict]) -> List['Category']:
        categories: List[Category] = []
        for value in values:
            categories.append(Category.from_dict(value))
        return categories

    @staticmethod
    def from_dict(value: dict) -> 'Category':
        return Category.parse_obj(value)


def create_category(id: str, name: str, type: CategoryType):
    return Category(id=id, name=name, description='', hotkey='', color='', type=type)



from learning_loop_node.trainer.tests.testing_trainer import TestingTrainer
from learning_loop_node.trainer.tests.states import state_helper
from learning_loop_node.trainer import active_training
from learning_loop_node.trainer.training import Training
import os


def create_detection_file(training: Training):
    active_training.detections.save(training, [])


async def test_cleanup_successfull():
    state_helper.create_active_training_file(training_state='ready_for_cleanup')
    trainer = TestingTrainer()
    trainer.training = active_training.load()  # normally done by node
    create_detection_file(trainer.training)
    training = trainer.training
    active_training.detections_upload_progress.save(training, 42)

    assert trainer.training is not None
    assert active_training.exists() is True
    assert active_training.detections.exists(training) is True
    assert active_training.detections_upload_progress.exists(training) is True

    await trainer.clear_training()

    assert trainer.training is None
    assert active_training.exists() is False
    assert active_training.detections.exists(training) is False
    assert active_training.detections_upload_progress.exists(training) is False



from abc import abstractmethod
import asyncio
import os
from typing import Dict, List, Optional, Union
from uuid import uuid4
from learning_loop_node.rest.downloader import DataDownloader
from learning_loop_node.loop import loop
from tqdm import tqdm
from ..model_information import ModelInformation
from .executor import Executor
from .training import Training
from .model import BasicModel, Model, PretrainedModel
from ..context import Context
from ..node import Node
from .downloader import TrainingsDownloader
from ..rest import downloads, uploads
from .. import node_helper
import logging
from .helper import is_valid_uuid4
from glob import glob
import json
from fastapi.encoders import jsonable_encoder
import shutil
from learning_loop_node.data_classes.category import Category
from learning_loop_node.trainer.hyperparameter import Hyperparameter
import time
from time import perf_counter
from learning_loop_node.trainer.training import State as TrainingState
from learning_loop_node.trainer.training_data import TrainingData
from learning_loop_node.trainer import active_training
from learning_loop_node.trainer import training_syncronizer
import socketio
from datetime import datetime
from .errors import TrainingError
from .errors import Errors


class Trainer():

    def __init__(self, model_format: str) -> None:
        self.model_format: str = model_format
        self.training: Optional[Training] = None
        self.executor: Optional[Executor] = None
        self.start_time: Optional[int] = None
        self.training_task: Optional[asyncio.Task] = None
        self.train_task: Optional[asyncio.Task] = None
        self.errors = Errors()
        self.shutdown_event: asyncio.Event = asyncio.Event()

    def init(self, context: Context, details: dict) -> None:
        try:
            self.training = Trainer.generate_training(context)
            self.training.data = TrainingData(categories=Category.from_list(details['categories']))
            self.training.data.hyperparameter = Hyperparameter.from_dict(details)
            self.training.training_number = details['training_number']
            self.training.base_model_id = details['id']
            self.training.training_state = TrainingState.Initialized
            active_training.save(self.training)
            logging.info(f'init training: {self.training}')
        except:
            logging.exception('Error in init')

    async def train(self, uuid, sio_client) -> None:
        self.start_time = time.time()
        self.errors.reset_all()
        try:
            self.training_task = asyncio.get_running_loop().create_task(self._train(uuid, sio_client))
            await self.training_task

        except asyncio.CancelledError as e:
            if not self.shutdown_event.is_set():
                logging.error(str(e))

                logging.info('cancelled training task')
                self.training.training_state = TrainingState.ReadyForCleanup
                active_training.save(self.training)
                await self.clear_training()

        except BaseException:
            logging.exception('Error in train')
        finally:
            self.start_time = None

    async def _train(self, uuid, sio_client) -> None:
        training = None
        if self.training:
            training = self.training
        elif active_training.exists():
            logging.warning('found active training on hd')
            training = active_training.load()
            logging.warning(jsonable_encoder(training))
            self.training = training

        while self.training or active_training.exists():
            if training.training_state == TrainingState.Initialized:
                await self.prepare()
            if training.training_state == TrainingState.DataDownloaded:
                await self.download_model()
            if training.training_state == TrainingState.TrainModelDownloaded:
                await self.run_training(uuid, sio_client)
            if training.training_state == TrainingState.TrainingFinished:
                await self.ensure_confusion_matrix_synced(uuid, sio_client)
            if training.training_state == TrainingState.ConfusionMatrixSynced:
                await self.upload_model()
            if training.training_state == TrainingState.TrainModelUploaded:
                await self.do_detections()
            if training.training_state == TrainingState.Detected:
                await self.upload_detections()
            if training.training_state == TrainingState.ReadyForCleanup:
                await self.clear_training()
            else:
                await asyncio.sleep(1)

    async def prepare(self) -> None:
        previous_state = self.training.training_state
        self.training.training_state = TrainingState.DataDownloading
        error_key = 'prepare'
        try:
            await self._prepare()
        except asyncio.CancelledError:
            raise
        except Exception as e:
            logging.exception("Unknown error in 'prepare'")
            self.training.training_state = previous_state
            self.errors.set(error_key, str(e))
        else:
            self.errors.reset(error_key)
            self.training.training_state = TrainingState.DataDownloaded
            active_training.save(self.training)

    async def _prepare(self) -> None:
        downloader = TrainingsDownloader(self.training.context)
        image_data, skipped_image_count = await downloader.download_training_data(self.training.images_folder)
        self.training.data.image_data = image_data
        self.training.data.skipped_image_count = skipped_image_count

    async def download_model(self) -> None:
        previous_state = self.training.training_state
        self.training.training_state = TrainingState.TrainModelDownloading
        error_key = 'download_model'
        try:
            await self._download_model()
        except asyncio.CancelledError:
            raise
        except Exception as e:
            logging.exception('download_model failed')
            self.training.training_state = previous_state
            self.errors.set(error_key, str(e))
        else:
            self.errors.reset(error_key)
            logging.info('download_model_task finished')
            self.training.training_state = TrainingState.TrainModelDownloaded
            active_training.save(self.training)

    async def _download_model(self) -> None:
        model_id = self.training.base_model_id
        if is_valid_uuid4(self.training.base_model_id):
            logging.debug('loading model from Learning Loop')
            logging.info(f'downloading model {model_id} as {self.model_format}')
            await downloads.download_model(self.training.training_folder, self.training.context, model_id, self.model_format)
            shutil.move(f'{self.training.training_folder}/model.json',
                        f'{self.training.training_folder}/base_model.json')

    async def run_training(self, trainer_node_uuid: str, sio_client: socketio.AsyncClient) -> None:
        error_key = 'run_training'
        # NOTE normally we reset errors after the step was successful. We do not want to display an old error during the whole training.
        self.errors.reset(error_key)
        previous_state = self.training.training_state
        try:
            self.executor = Executor(self.training.training_folder)
            self.training.training_state = TrainingState.TrainingRunning
            self.train_task = None

            if self.can_resume():
                self.train_task = self.resume()
            else:
                model_id = self.training.base_model_id
                if not is_valid_uuid4(model_id):
                    self.train_task = self.start_training_from_scratch(model_id)
                else:
                    self.train_task = self.start_training()

            await self.train_task

            last_sync_time = datetime.now()
            while True:
                if not self.executor.is_process_running():
                    break
                if (datetime.now() - last_sync_time).total_seconds() > 5:
                    last_sync_time = datetime.now()
                    error = self.get_error()
                    if error:
                        self.errors.set(error_key, error)
                    else:
                        self.errors.reset(error_key)

                    try:
                        await self.sync_confusion_matrix(trainer_node_uuid, sio_client)
                    except asyncio.CancelledError:
                        raise
                    except:
                        pass
                else:
                    await asyncio.sleep(0.1)

            error = self.get_error()
            if error:
                self.errors.set(error_key, error)
                raise TrainingError(cause=error)
            else:
                self.errors.reset(error_key)

        except asyncio.CancelledError:
            raise
        except TrainingError as e:
            logging.exception('Error in TrainingProcess')
            self.training.training_state = previous_state
        except Exception as e:
            self.errors.set(error_key, f'Could not start training {str(e)}')
            self.training.training_state = previous_state
            logging.exception('Error in run_training')
        else:
            self.training.training_state = TrainingState.TrainingFinished
            active_training.save(self.training)

    async def ensure_confusion_matrix_synced(self, trainer_node_uuid: str, sio_client: socketio.AsyncClient):
        previous_state = self.training.training_state
        self.training.training_state = TrainingState.ConfusionMatrixSyncing
        try:
            await self.sync_confusion_matrix(trainer_node_uuid, sio_client)
        except asyncio.CancelledError:
            raise
        except Exception as e:
            logging.exception('Error in ensure_confusion_matrix_synced')
            self.training.training_state = previous_state
        else:
            self.training.training_state = TrainingState.ConfusionMatrixSynced
            active_training.save(self.training)

    async def sync_confusion_matrix(self, trainer_node_uuid: str, sio_client: socketio.AsyncClient):
        error_key = 'sync_confusion_matrix'
        try:
            await training_syncronizer.try_sync_model(self, trainer_node_uuid, sio_client)
        except socketio.exceptions.BadNamespaceError as e:
            logging.error('Error during confusion matrix syncronization. BadNamespaceError')
            self.errors.set(error_key, str(e))
            raise
        except Exception as e:
            logging.exception('Error during confusion matrix syncronization')
            self.errors.set(error_key, str(e))
            raise
        else:
            self.errors.reset(error_key)

    async def upload_model(self) -> None:
        error_key = 'upload_model'
        previous_state = self.training.training_state
        self.training.training_state = TrainingState.TrainModelUploading

        try:
            uploaded_model = await self._upload_model(self.training.context)
            self.training.model_id_for_detecting = uploaded_model['id']
        except asyncio.CancelledError:
            raise
        except Exception as e:
            logging.exception('Error in upload_model')
            self.errors.set(error_key, str(e))
            self.training.training_state = previous_state
        else:
            self.errors.reset(error_key)
            self.training.training_state = TrainingState.TrainModelUploaded
            active_training.save(self.training)

    async def _upload_model(self, context: Context) -> dict:
        try:
            files = await asyncio.get_running_loop().run_in_executor(None, self.get_latest_model_files)
        except FileNotFoundError as e:
            raise Exception('Could not find any model files to upload.')

        model_json_content = self.create_model_json_content()
        model_json_path = '/tmp/model.json'
        with open(model_json_path, 'w') as f:
            json.dump(model_json_content, f)

        if isinstance(files, list):
            files = {self.model_format: files}
        uploaded_model = None
        already_uploaded_formats = active_training.model_upload_progress.load(self.training)
        if isinstance(files, dict):
            for format in files:
                if format in already_uploaded_formats:
                    continue
                # model.json was mandatory in previous versions. Now its forbidden to provide an own model.json file.
                assert len([file for file in files[format] if 'model.json' in file]) == 0, \
                    "It is not allowed to provide a 'model.json' file."
                _files = files[format]
                _files.append(model_json_path)
                uploaded_model = await uploads.upload_model_for_training(context, _files, self.training.training_number, format)
                already_uploaded_formats.append(format)
                active_training.model_upload_progress.save(self.training, already_uploaded_formats)

        else:
            raise TypeError(f'can only save model as list or dict, but was {files}')
        return uploaded_model

    async def do_detections(self):
        error_key = 'detecting'
        previous_state = self.training.training_state
        try:
            self.training.training_state = TrainingState.Detecting
            detections = await self._do_detections()
            active_training.detections.save(self.training, jsonable_encoder(detections))
        except asyncio.CancelledError:
            raise
        except Exception as e:
            self.errors.set(error_key, str(e))
            logging.exception('Error in do_detections')
            self.training.training_state = previous_state
        else:
            self.errors.reset(error_key)
            self.training.training_state = TrainingState.Detected
            active_training.save(self.training)

    async def _do_detections(self) -> List:
        context = self.training.context
        model_id = self.training.model_id_for_detecting
        tmp_folder = f'/tmp/model_for_auto_detections_{model_id}_{self.model_format}'

        shutil.rmtree(tmp_folder, ignore_errors=True)
        os.makedirs(tmp_folder)
        logging.info('downloading model for detecting')

        await downloads.download_model(tmp_folder, context, model_id, self.model_format)
        with open(f'{tmp_folder}/model.json', 'r') as f:
            content = json.load(f)
            model_information = ModelInformation.parse_obj(content)

        project_folder = Node.create_project_folder(context)
        image_folder = node_helper.create_image_folder(project_folder)
        downloader = DataDownloader(context)
        image_ids = []
        for state in ['inbox', 'annotate', 'review', 'complete']:
            logging.info(f'fetching image ids of {state}')
            new_ids = await downloader.fetch_image_ids(query_params=f'state={state}')
            image_ids += new_ids
            logging.info(f'downloading {len(new_ids)} images')
            await downloader.download_images(new_ids, image_folder)
        images = await asyncio.get_event_loop().run_in_executor(None, Trainer.images_for_ids, image_ids, image_folder)
        logging.info(f'running detections on {len(images)} images')
        detections = await self._detect(model_information, images, tmp_folder)
        return detections

    async def upload_detections(self):
        error_key = 'upload_detections'
        previous_state = self.training.training_state
        self.training.training_state = TrainingState.DetectionUploading
        context = self.training.context

        await asyncio.sleep(0.1)  # NOTE needed for tests

        try:
            detections = active_training.detections.load(self.training)
            await self._upload_detections(context, detections)
        except asyncio.CancelledError:
            raise
        except Exception as e:
            self.errors.set(error_key, str(e))
            logging.exception('Error in upload_detections')
            self.training.training_state = previous_state
        else:
            self.errors.reset(error_key)
            self.training.training_state = TrainingState.ReadyForCleanup
            active_training.save(self.training)

    async def _upload_detections(self, context: Context, detections: List[dict]):
        logging.info('uploading detections')
        batch_size = 500

        skip_detections = active_training.detections_upload_progress.load(self.training)

        for i in tqdm(range(skip_detections, len(detections), batch_size), position=0, leave=True):
            batch_detections = detections[i:i+batch_size]
            logging.info(f'uploading detections. File size : {len(json.dumps(batch_detections))}')

            async with loop.post(f'api/{context.organization}/projects/{context.project}/detections', json=batch_detections) as response:
                if response.status != 200:
                    msg = f'could not upload detections. {str(response)}'
                    logging.error(msg)
                    raise Exception(msg)
                else:
                    logging.info('successfully uploaded detections')
                    active_training.detections_upload_progress.save(self.training, min(i+batch_size, len(detections)))

    async def clear_training(self):
        active_training.detections.delete(self.training)
        active_training.detections_upload_progress.delete(self.training)
        try:
            await self.clear_training_data(self.training.training_folder)
        except NotImplementedError:
            logging.exception('clear_training_data not implemented')
            pass
        else:
            self.training = None
            active_training.delete()

    def stop(self) -> None:
        if not self.training:
            return
        if self.executor and self.executor.is_process_running():
            self.executor.stop()
        elif self.training_task:
            logging.error('cancelling training task')
            self.training_task.cancel()

    def shutdown(self) -> None:
        self.shutdown_event.set()
        self.stop()
        self.stop()  # NOTE first stop may only stop training.

    async def start_training(self) -> None:
        raise NotImplementedError()

    async def start_training_from_scratch(self, identifier: str) -> None:
        raise NotImplementedError()

    def can_resume(self) -> bool:
        '''Override this method to return True if the trainer can resume training.'''
        return False

    async def resume(self) -> None:
        '''Is called when self.can_resume() returns True.
           One may resume the training on a previously trained model stored by self.on_model_published(basic_model).
        '''
        raise NotImplementedError()

    def get_error(self) -> Optional[Union[None, str]]:
        '''Should be used to provide error informations to the Learning Loop by extracting data from self.executor.get_log().'''
        pass

    def get_log(self) -> str:
        return self.executor.get_log()

    def get_new_model(self) -> Optional[BasicModel]:
        '''Is called frequently to check if a new "best" model is availabe.
        Returns None if no new model could be found. Otherwise BasicModel(confusion_matrix, meta_information).
        `confusion_matrix` contains a dict of all classes:
            - The classes must be identified by their id, not their name.
            - For each class a dict with tp, fp, fn is provided (true positives, false positives, false negatives).
        `meta_information` can hold any data which is helpful for self.on_model_published to store weight file etc for later upload via self.get_model_files
        '''
        raise NotImplementedError()

    def on_model_published(self, basic_model: BasicModel) -> None:
        '''Called after a BasicModel has been successfully send to the Learning Loop.
        The files for this model should be stored.
        self.get_latest_model_files is used to gather all files needed for transfering the actual data from the trainer node to the Learning Loop.
        In the simplest implementation this method just renames the weight file (encoded in BasicModel.meta_information) into a file name like latest_published_model
        '''
        raise NotImplementedError()

    def get_latest_model_files(self) -> Union[List[str], Dict[str, List[str]]]:
        '''Called when the Learning Loop requests to backup the latest model for the training.
        Should return a list of file paths which describe the model.
        These files must contain all data neccessary for the trainer to resume a training (eg. weight file, hyperparameters, etc.)
        and will be stored in the Learning Loop unter the format of this trainer.
        Note: by convention the weightfile should be named "model.<extension>" where extension is the file format of the weightfile.
        For example "model.pt" for pytorch or "model.weights" for darknet/yolo.

        If a trainer can also generate other formats (for example for an detector),
        a dictionary mapping format -> list of files can be returned.
        '''
        raise NotImplementedError()

    @staticmethod
    def images_for_ids(image_ids, image_folder) -> List[str]:
        logging.info(f'### Going to get images for {len(image_ids)} images ids')
        start = perf_counter()
        images = [img for img in glob(f'{image_folder}/**/*.*', recursive=True)
                  if os.path.splitext(os.path.basename(img))[0] in image_ids]
        end = perf_counter()
        logging.info(f'found {len(images)} images for {len(image_ids)} image ids, which took {end-start:0.2f} seconds')
        return images

    async def _detect(self, model_information: ModelInformation, images:  List[str], model_folder: str) -> List:
        raise NotImplementedError()

    async def clear_training_data(self, training_folder: str) -> None:
        '''Called after a training has finished. Deletes all data that is not needed anymore after a training run. This can be old
        weightfiles or any additional files.
        '''
        raise NotImplementedError()

    @property
    @abstractmethod
    def provided_pretrained_models(self) -> List[PretrainedModel]:
        raise NotImplementedError()

    @staticmethod
    def generate_training(context: Context) -> Training:
        training_uuid = str(uuid4())
        project_folder = Node.create_project_folder(context)
        return Training(
            id=training_uuid,
            context=context,
            project_folder=project_folder,
            images_folder=node_helper.create_image_folder(project_folder),
            training_folder=Trainer.create_training_folder(project_folder, training_uuid)
        )

    @staticmethod
    def create_training_folder(project_folder: str, trainings_id: str) -> str:
        training_folder = f'{project_folder}/trainings/{trainings_id}'
        os.makedirs(training_folder, exist_ok=True)
        return training_folder

    @property
    def hyperparameters(self) -> dict:
        if self.training and self.training.data:
            information = {}
            information['resolution'] = self.training.data.hyperparameter.resolution
            information['flipRl'] = self.training.data.hyperparameter.flip_rl
            information['flipUd'] = self.training.data.hyperparameter.flip_ud
            return information
        else:
            return None

    @property
    def model_architecture(self) -> Union[str, None]:
        return None

    def create_model_json_content(self):
        content = {
            'categories': [c.dict() for c in self.training.data.categories],
            'resolution': self.training.data.hyperparameter.resolution
        }
        return content



import multiprocessing
from shutil import ExecError
from learning_loop_node.trainer.tests.testing_trainer import TestingTrainer
from learning_loop_node.trainer.trainer_node import TrainerNode
import pytest
from learning_loop_node import DetectorNode, ModelInformation
from learning_loop_node.detector import Outbox
from learning_loop_node.data_classes import Category

import uvicorn
from multiprocessing import Process, log_to_stderr
import logging
import icecream
import os
import socket
import asyncio
from typing import Generator
import socketio
from glob import glob

logging.basicConfig(level=logging.INFO)

# show ouptut from uvicorn server https://stackoverflow.com/a/66132186/364388
log_to_stderr(logging.INFO)

icecream.install()


def pytest_configure():
    pytest.trainer_port = 5001


@pytest.fixture()
async def test_trainer_node(request):
    os.environ['ORGANIZATION'] = 'zauberzeug'
    os.environ['PROJECT'] = 'demo'

    trainer = TestingTrainer()

    node = TrainerNode(name='test', trainer=trainer, uuid='00000000-0000-0000-0000-000000000000')
    await port_is(free=True)

    multiprocessing.set_start_method('fork', force=True)
    assert multiprocessing.get_start_method() == 'fork'
    proc = Process(target=uvicorn.run,
                   args=(node,),
                   kwargs={
                       "host": "127.0.0.1",
                       "port": pytest.trainer_port,

                   }, daemon=True)
    proc.start()
    await port_is(free=False)
    yield node

    try:
        await node.shutdown()
    except:
        logging.exception('error while shutting down node')

    try:
        proc.kill()
    except:  # for python 3.6
        proc.terminate()
    proc.join()


def is_port_in_use(port):
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        return s.connect_ex(('localhost', port)) == 0


async def port_is(free: bool):
    for i in range(10):
        if not free and is_port_in_use(pytest.trainer_port):
            return
        if free and not is_port_in_use(pytest.trainer_port):
            return
        else:
            await asyncio.sleep(0.5)
    raise Exception(f'port {pytest.trainer_port} is {"not" if free else ""} free')


@pytest.fixture(autouse=True, scope='session')
def initialize_active_training():
    from learning_loop_node.trainer import active_training
    active_training.init('00000000-0000-0000-0000-000000000000')
    yield




from learning_loop_node.detector.detections import Detections
from learning_loop_node.detector.detector_node import DetectorNode
from learning_loop_node.detector.outbox.outbox import Outbox
from PIL import Image
import os
import numpy as np
import pytest
import shutil
from multiprocessing import Event


@pytest.fixture()
def outbox():
    os.environ['LOOP_ORGANIZATION'] = 'zauberzeug'
    os.environ['LOOP_PROJECT'] = 'demo'
    outbox = Outbox()
    shutil.rmtree(outbox.path, ignore_errors=True)
    os.mkdir(outbox.path)

    yield outbox
    shutil.rmtree(outbox.path, ignore_errors=True)


def test_files_are_deleted_after_sending(outbox: Outbox):
    assert outbox.path.startswith('/tmp')
    os.mkdir(f'{outbox.path}/test')
    with open(f'{outbox.path}/test/image.json', 'w') as f:
        f.write('{"box_detections":[]}')
        f.close()

    img = Image.new('RGB', (60, 30), color=(73, 109, 137))
    img.save(f'{outbox.path}/test/image.jpg')

    items = outbox.get_data_files()
    assert len(items) == 1

    outbox.upload()

    items = outbox.get_data_files()
    assert len(items) == 0


def test_saving_opencv_image(outbox: Outbox):
    img = np.ones((300, 300, 1), np.uint8)*255
    outbox.save(img)
    items = outbox.get_data_files()
    assert len(items) == 1


def test_saving_binary(outbox: Outbox):
    assert len(outbox.get_data_files()) == 0
    img = Image.new('RGB', (60, 30), color=(73, 109, 137))
    img.save('/tmp/image.jpg')
    with open(f'/tmp/image.jpg', 'rb') as f:
        data = f.read()
    outbox.save(data)
    assert len(outbox.get_data_files()) == 1


def test_files_are_automatically_uploaded(test_detector_node: DetectorNode):
    test_detector_node.outbox.save(Image.new('RGB', (60, 30), color=(73, 109, 137)).tobytes(), Detections())
    assert len(test_detector_node.outbox.get_data_files()) == 1

    assert len(test_detector_node.outbox.get_data_files()) == 1



import asyncio
from datetime import datetime, timedelta
from async_generator import asynccontextmanager
import aiohttp
import os
import werkzeug
from icecream import ic
import logging
import requests
from . import environment_reader


class AccessToken():
    def __init__(self, token: str, local_expire_date: datetime):
        self.token = token
        self.local_expire_date = local_expire_date

    def is_still_valid(self) -> bool:
        return (self.local_expire_date - timedelta(hours=1)) > datetime.now()

    def is_invalid(self) -> bool:
        return not self.is_still_valid


def token_from_response(response: werkzeug.Response) -> AccessToken:
    content = response.json()
    server_time = werkzeug.http.parse_date(response.headers['date'])
    expires_time = datetime.fromtimestamp(content['expires'], tz=server_time.tzinfo)
    # we do not care about the few seconds wich ellapsed during request.
    local_expire_time = datetime.now() + (expires_time - server_time)
    return AccessToken(content['access_token'], local_expire_time)


class Loop():
    def __init__(self) -> None:
        host = os.environ.get('LOOP_HOST', None) or os.environ.get('HOST', 'learning-loop.ai')
        self.base_url: str = f'http{"s" if host != "backend" else ""}://' + host
        self.username: str = os.environ.get('LOOP_USERNAME', None) or os.environ.get('USERNAME', None)
        self.password: str = os.environ.get('LOOP_PASSWORD', None) or os.environ.get('PASSWORD', None)
        self.organization: str = environment_reader.organization(default='')
        self.project: str = environment_reader.project(default='')
        self.access_token = None
        self.web = requests.Session()

    def download_token(self):

        response = self.web.post(
            (self.base_url + '/api/token').replace('//api', '/api'),
            data={'username': self.username, 'password': self.password}
        )
        response.raise_for_status()
        return token_from_response(response)

    async def create_headers(self) -> dict:
        return await asyncio.get_event_loop().run_in_executor(None, self.get_headers)

    def get_headers(self):
        headers = {}
        if self.username and self.password:
            if self.access_token is None or self.access_token.is_invalid():
                self.access_token = self.download_token()

            headers['Authorization'] = f'Bearer {self.access_token.token}'

        return headers

    @asynccontextmanager
    async def get(self, path):
        url = f'{self.base_url}/{path.lstrip("/")}'
        logging.debug(url)
        async with aiohttp.ClientSession(headers=await self.create_headers()) as session:
            async with session.get(url) as response:
                yield response

    async def get_json_async(self, path):
        url = f'{loop.project_path}{path}'
        async with self.get(url) as response:
            if response.status != 200:
                raise Exception('bad response: ' + str(response))
            return await response.json()

    def get_json(self, path):
        return asyncio.get_event_loop().run_until_complete(self.get_json_async(path))

    async def get_data_async(self, path):
        async with self.get(f'{loop.project_path}{path}') as response:
            if response.status != 200:
                raise Exception('bad response: ' + str(response))
            return await response.read()

    def get_data(self, path):
        return asyncio.get_event_loop().run_until_complete(self.get_data_async(path))

    @asynccontextmanager
    async def put(self, path, data):
        async with aiohttp.ClientSession(headers=await self.create_headers()) as session:
            async with session.put(f'{self.base_url}/{path}', data=data) as response:
                yield response

    @asynccontextmanager
    async def post(self, path, **kwargs):
        async with aiohttp.ClientSession(headers=await self.create_headers()) as session:
            async with session.post(f'{self.base_url}/{path}', **kwargs) as response:
                yield response

    @property
    def project_path(self):
        return f'/api/{self.organization}/projects/{self.project}'


loop = Loop()




from learning_loop_node.node import Node
from fastapi import APIRouter,  Request, HTTPException
from fastapi.responses import PlainTextResponse
import logging
from enum import Enum

router = APIRouter()


class OperationMode(str, Enum):
    Startup = 'startup'  # used until model is loaded
    Idle = 'idle'  # will check and perform updates
    Detecting = 'detecting'  # Blocks updates


@router.put("/operation_mode")
async def operation_mode(request: Request):
    '''
    Example Usage
        curl -X PUT -d "check_for_updates" http://localhost/operation_mode
        curl -X PUT -d "detecting" http://localhost/operation_mode
    '''

    content = str(await request.body(), 'utf-8')
    try:
        target_mode = OperationMode(content)
    except ValueError as e:
        raise HTTPException(422, str(e))

    node: Node = request.app

    logging.info(f'current node state : {node.status.state}')
    logging.info(f'target operation mode : {target_mode}')

    await node.set_operation_mode(target_mode)
    return "OK"


@router.get("/operation_mode")
async def operation_mode(request: Request):
    '''
    Example Usage
        curl http://localhost/operation_mode
    '''
    return PlainTextResponse(request.app.operation_mode.value)



import os
from glob import glob
from threading import Thread
from fastapi.encoders import jsonable_encoder
import os
from learning_loop_node.detector import Detections
from typing import List
import json
from datetime import datetime
from ...globals import GLOBALS
from ... import environment_reader
from multiprocessing import Event
import time
import logging
import requests
import shutil


class Outbox():

    def __init__(self) -> None:
        self.path = f'{GLOBALS.data_folder}/outbox'
        if not os.path.exists(self.path):
            os.makedirs(self.path)

        host = os.environ.get('HOST', 'learning-loop.ai')
        base: str = f'http{"s" if host != "backend" else ""}://' + host
        o = environment_reader.organization()
        p = environment_reader.project()
        assert o and p, 'Outbox needs an organization and a projekct '
        self.target_uri = f'{base}/api/{o}/projects/{p}/images'

        self.log = logging.getLogger()
        self.shutdown_event = None
        self.upload_thread = None

    def save(self, image: bytes, detections: Detections = Detections(), tags: List[str] = []) -> None:
        id = datetime.now().isoformat(sep='_', timespec='milliseconds')
        tmp = f'{GLOBALS.data_folder}/tmp/{id}'
        detections.tags = tags
        detections.date = id
        os.makedirs(tmp, exist_ok=True)
        with open(tmp + '/image.json', 'w') as f:
            json.dump(jsonable_encoder(detections), f)

        with open(tmp + '/image.jpg', 'wb') as f:
            f.write(image)
        os.rename(tmp, self.path + '/' + id)  # NOTE rename is atomic so upload can run in parallel

    def get_data_files(self):
        return glob(f'{self.path}/*')

    def start_continuous_upload(self):
        self.shutdown_event = Event()
        self.upload_thread = Thread(target=self._continuous_upload)
        self.upload_thread.start()

    def _continuous_upload(self):
        self.log.info('start continuous upload')
        while not self.shutdown_event.is_set():
            self.upload()
            time.sleep(1)
        self.log.info('stop continuous upload')

    def upload(self):
        items = self.get_data_files()
        self.log.info(f'Found {len(items)} images to upload')
        for item in items:
            if self.shutdown_event and self.shutdown_event.is_set():
                break
            try:
                data = [('files', open(f'{item}/image.json', 'r')),
                        ('files', open(f'{item}/image.jpg', 'rb'))]

                response = requests.post(self.target_uri, files=data)
                if response.status_code == 200:
                    shutil.rmtree(item)
                    self.log.info(f'uploaded {item} successfully')
                elif response.status_code == 422:
                    self.log.error(f'Broken content in {item}: dropping this data')
                    shutil.rmtree(item)
                else:
                    self.log.error(f'Could not upload {item}: {response.status_code}, {response.content}')
            except:
                self.log.exception('could not upload files')

    def stop_continuous_upload(self, timeout=5):
        try:
            self.shutdown_event.set()
            proc = self.upload_thread
            proc.join(timeout)
        except:
            pass
        if proc.is_alive():
            proc.terminate()
            self.log.info('terminated process')



from typing import List, Optional
import os
from glob import glob
import asyncio
from learning_loop_node.context import Context
from learning_loop_node.loop import loop
from learning_loop_node.rest import downloads
from learning_loop_node import node_helper
import logging
from icecream import ic
from time import perf_counter


class DataDownloader():
    context: Context

    def __init__(self, context: Context):
        self.context = context

    async def fetch_image_ids(self, query_params: Optional[str] = '') -> List[str]:
        async with loop.get(f'api/{self.context.organization}/projects/{self.context.project}/data?{query_params}') as response:
            assert response.status == 200, response
            return (await response.json())['image_ids']

    async def download_images_data(self, ids: List[str]) -> List[dict]:
        return await downloads.download_images_data(self.context.organization, self.context.project, ids)

    async def download_images(self, image_ids: List[str], image_folder: str) -> None:
        '''Will skip existing images'''
        new_image_ids = await asyncio.get_event_loop().run_in_executor(None, DataDownloader.filter_existing_images, image_ids, image_folder)
        paths, ids = node_helper.create_resource_paths(self.context.organization, self.context.project, new_image_ids)
        await downloads.download_images(paths, ids, image_folder)

    @staticmethod
    def filter_existing_images(all_image_ids, image_folder) -> List[str]:
        logging.info(f'### Going to filter {len(all_image_ids)} images ids')
        start = perf_counter()
        ids = [os.path.splitext(os.path.basename(image))[0]
               for image in glob(f'{image_folder}/*.jpg')]
        logging.info(f'found {len(ids)} images on disc')
        result = [id for id in all_image_ids if id not in ids]
        end = perf_counter()
        logging.info(f'calculated {len(result)} new image ids, which took {end-start:0.2f} seconds')
        return result



from enum import Enum
from learning_loop_node.context import Context
from learning_loop_node.trainer.downloader import TrainingsDownloader
from training import Training
from training import State as training_state


class SubState(str, Enum):
    Empty = "Empty"
    DataDownloaded = "DataDownloaded"


async def start(training: Training):
    if training.training_state != training_state.Preparing:
        raise Exception('Training is not in preparing state.')
    if training.training_sub_state is SubState.Empty:
        downloader = TrainingsDownloader(training.context))
        training.data=await downloader.download_training_data(training.images_folder)
        training.save()



from learning_loop_node.context import Context
from learning_loop_node.node import Node
from learning_loop_node.trainer.trainer import Trainer
from learning_loop_node import node_helper


def create_needed_folders(base_folder: str, training_uuid: str = 'some_uuid'):
    project_folder = Node.create_project_folder(
        Context(organization='zauberzeug', project='pytest', base_folder=base_folder))
    image_folder = node_helper.create_image_folder(project_folder)
    training_folder = Trainer.create_training_folder(project_folder, training_uuid)
    return project_folder, image_folder, training_folder



from learning_loop_node.trainer.tests.testing_trainer import TestingTrainer
import asyncio
from learning_loop_node.trainer.tests.states.state_helper import assert_training_state
from learning_loop_node.trainer.tests.states import state_helper
from learning_loop_node.trainer import active_training
from learning_loop_node.trainer.trainer import Trainer
from learning_loop_node.trainer.trainer_node import TrainerNode

error_key = 'sync_confusion_matrix'


def trainer_has_error(trainer: Trainer):
    return trainer.errors.has_error_for(error_key)


async def test_nothing_to_sync():
    state_helper.create_active_training_file(training_state='training_finished')
    trainer = TestingTrainer()
    trainer.training = active_training.load()  # normally done by node

    train_task = asyncio.get_running_loop().create_task(trainer.train(None, None))

    await assert_training_state(trainer.training, 'confusion_matrix_synced', timeout=1, interval=0.001)
    assert trainer_has_error(trainer) is False
    assert trainer.training.training_state == 'confusion_matrix_synced'
    assert active_training.load() == trainer.training


async def test_unsynced_model_available__sync_successful(test_trainer_node: TrainerNode, mocker):
    mock_socket_io_call(mocker, test_trainer_node, {'success': True})

    state_helper.create_active_training_file(training_state='training_finished')
    trainer = test_trainer_node.trainer
    trainer.training = active_training.load()  # normally done by node
    trainer.has_new_model = True

    train_task = asyncio.get_running_loop().create_task(trainer.train(
        uuid=test_trainer_node.uuid, sio_client=test_trainer_node.sio_client))
    await assert_training_state(trainer.training, 'confusion_matrix_synced', timeout=1, interval=0.001)

    assert trainer_has_error(trainer) is False
    assert trainer.training.training_state == 'confusion_matrix_synced'
    assert active_training.load() == trainer.training


async def test_unsynced_model_available__sio_not_connected(test_trainer_node: TrainerNode):
    state_helper.create_active_training_file(training_state='training_finished')
    trainer = test_trainer_node.trainer
    trainer.training = active_training.load()  # normally done by node

    assert test_trainer_node.sio_client.connected is False
    trainer.has_new_model = True
    train_task = asyncio.get_running_loop().create_task(trainer.train(
        uuid=test_trainer_node.uuid, sio_client=test_trainer_node.sio_client))

    await assert_training_state(trainer.training, 'confusion_matrix_syncing', timeout=1, interval=0.001)
    await assert_training_state(trainer.training, 'training_finished', timeout=1, interval=0.001)

    assert trainer_has_error(trainer)
    assert trainer.training.training_state == 'training_finished'
    assert active_training.load() == trainer.training


async def test_unsynced_model_available__request_is_not_successful(test_trainer_node: TrainerNode, mocker):
    mock_socket_io_call(mocker, test_trainer_node, {'success': False})

    state_helper.create_active_training_file(training_state='training_finished')
    trainer = test_trainer_node.trainer
    trainer.training = active_training.load()  # normally done by node

    trainer.has_new_model = True
    train_task = asyncio.get_running_loop().create_task(trainer.train(
        uuid=test_trainer_node.uuid, sio_client=test_trainer_node.sio_client))

    await assert_training_state(trainer.training, 'confusion_matrix_syncing', timeout=1, interval=0.001)
    await assert_training_state(trainer.training, 'training_finished', timeout=1, interval=0.001)

    assert trainer_has_error(trainer)
    assert trainer.training.training_state == 'training_finished'
    assert active_training.load() == trainer.training


async def test_basic_mock(test_trainer_node: TrainerNode, mocker):
    patched_call_return_value = asyncio.Future()
    patched_call_return_value.set_result(
        {'success': True})
    mocker.patch.object(test_trainer_node.sio_client, 'call', return_value=patched_call_return_value)
    assert await test_trainer_node.sio_client.call() == {'success': True}


def mock_socket_io_call(mocker, trainer_node: TrainerNode, return_value):
    patched_call_return_value = asyncio.Future()
    patched_call_return_value.set_result(
        return_value)
    mocker.patch.object(trainer_node.sio_client, 'call', return_value=patched_call_return_value)



from learning_loop_node.context import Context
import logging
from learning_loop_node.globals import GLOBALS
from .status import Status, State
from fastapi import FastAPI
import socketio
import asyncio
import asyncio
import os
from icecream import ic
from .loop import loop
from fastapi_utils.tasks import repeat_every
import logging
from uuid import uuid4
from .socket_response import ensure_socket_response
from datetime import datetime
from . import log_conf
import json

log_conf.init()


class Node(FastAPI):
    name: str
    uuid: str

    def __init__(self, name: str, uuid: str = None):
        super().__init__()
        self.log = logging.getLogger()

        host = os.environ.get('LOOP_HOST', None) or os.environ.get('HOST', 'learning-loop.ai')
        self.ws_url = f'ws{"s" if host != "backend" else ""}://' + host

        self.name = name
        self.uuid = self.read_or_create_uuid(self.name) if uuid is None else uuid
        self.startup_time = datetime.now()

        self.sio_client = socketio.AsyncClient(
            reconnection_delay=0,
            request_timeout=0.5,
            # logger=True, engineio_logger=True
        )
        self.sio_client._trigger_event = ensure_socket_response(self.sio_client._trigger_event)
        self.reset()

        @self.sio_client.on('connect')
        async def on_connect():
            logging.debug('received "on_connect" from constructor event.')
            self.reset()
            state = self.get_state()
            try:
                await self.update_state(state)
            except:
                logging.exception(f'Error sending state.')
                raise

        @self.sio_client.on('disconnect')
        async def on_disconnect():
            logging.debug('received "on_disconnect" from constructor event.')
            await self.update_state(State.Offline)

        self.register_lifecycle_events()

    def read_or_create_uuid(self, identifier: str) -> str:
        identifier = identifier.lower().replace(' ', '_')
        uuids = {}
        file_path = f'{GLOBALS.data_folder}/uuids.json'
        if os.path.exists(file_path):
            with open(file_path, 'r') as f:
                uuids = json.load(f)

        uuid = uuids.get(identifier, None)
        if not uuid:
            uuid = str(uuid4())
            uuids[identifier] = uuid
            with open(file_path, 'w') as f:
                json.dump(uuids, f)
        return uuid

    def register_lifecycle_events(self):
        @self.on_event("startup")
        async def startup():
            logging.debug('received "startup" event')
            Node._activate_asyncio_warnings()

        @self.on_event("shutdown")
        async def shutdown():
            logging.debug('received "shutdown" event')
            await self.sio_client.disconnect()

        @self.on_event("startup")
        @repeat_every(seconds=10, raise_exceptions=False, wait_first=False)
        async def ensure_connected() -> None:
            logging.info(f'###732 current connection state: {self.sio_client.connected}')
            if not self.sio_client.connected:
                await self.connect()

    def reset(self):
        self.status = Status(id=self.uuid, name=self.name)

    async def get_state(self):
        raise Exception("Override this in subclass")

    async def connect(self):
        try:
            await self.sio_client.disconnect()
        except:
            pass

        logging.info(f'connecting to Learning Loop at {self.ws_url}')
        try:
            headers = await self.get_sio_headers()
            await self.sio_client.connect(f"{self.ws_url}", headers=headers, socketio_path="/ws/socket.io")
            logging.debug(f'my sid is {self.sio_client.sid}')
            logging.debug(f"connecting as type {headers['nodeType']}")
            logging.info(f'connected to Learning Loop at {self.ws_url}')
        except socketio.exceptions.ConnectionError as e:
            logging.error(f'socket.io connection error to "{self.ws_url}"')
        except Exception:
            logging.exception(f'error while connecting to "{self.ws_url}"')

    async def update_state(self, state: State):
        self.status.state = state
        if self.status.state != State.Offline:
            await self.send_status()

    async def send_status(self):
        raise Exception("Override this in subclass")

    async def get_sio_headers(self) -> dict:
        headers = await asyncio.get_event_loop().run_in_executor(None, loop.get_headers)
        headers['organization'] = loop.organization
        headers['project'] = loop.project
        headers['nodeType'] = self.get_node_type()
        return headers

    def get_node_type(self):
        classname = self.__class__.__name__
        if classname == 'TrainerNode':
            return 'trainer'
        elif classname == 'DetectorNode':
            return 'detector'
        elif classname == 'ConverterNode':
            return 'converter'
        elif classname == 'AnnotationNode':
            return 'annotation_node'
        else:
            return classname

    @staticmethod
    def create_project_folder(context: Context) -> str:
        project_folder = f'{GLOBALS.data_folder}/{context.organization}/{context.project}'
        os.makedirs(project_folder, exist_ok=True)
        return project_folder

    @staticmethod
    def _activate_asyncio_warnings() -> None:
        '''Produce warnings for coroutines which take too long on the main loop and hence clog the event loop'''
        try:
            import sys
            if sys.version_info.major >= 3 and sys.version_info.minor >= 7:  # most
                loop = asyncio.get_running_loop()
            else:
                loop = asyncio.get_event_loop()

            loop.set_debug(True)
            loop.slow_callback_duration = 0.2
            logging.info('activated asyncio warnings')
        except:
            logging.exception('could not activate asyncio warnings')



from typing import List, Tuple
import os


def create_resource_paths(organization_name: str, project_name: str, image_ids: List[str]) -> Tuple[List[str], List[str]]:
    if not image_ids:
        return [], []
    url_ids = [(f'api/{organization_name}/projects/{project_name}/images/{id}/main', id)
               for id in image_ids]
    urls, ids = list(map(list, zip(*url_ids)))

    return urls, ids


def create_image_folder(project_folder: str) -> str:
    image_folder = f'{project_folder}/images'
    os.makedirs(image_folder, exist_ok=True)
    return image_folder



#!/usr/bin/env python3
import requests
from icecream import ic
# https://stackoverflow.com/a/39225272/4082686


def g_download(id, destination):
    URL = "https://docs.google.com/uc?export=download"

    session = requests.Session()

    response = session.get(URL, params={
        'id': id,
        'confirm': 't'},  # e.g. large file warning.
        stream=True)
    ic(str(response))
    token = get_confirm_token(response)

    if token:
        params = {'id': id, 'confirm': token}
        response = session.get(URL, params=params, stream=True)

    save_response_content(response, destination)


def get_confirm_token(response):
    for key, value in response.cookies.items():
        if key.startswith('download_warning'):
            return value

    return None


def save_response_content(response, destination):
    CHUNK_SIZE = 32768

    with open(destination, "wb") as f:
        for chunk in response.iter_content(CHUNK_SIZE):
            if chunk:  # filter out keep-alive new chunks
                f.write(chunk)


if __name__ == "__main__":
    file_id = '1q8nT-CTHt1eZuNjPMbdaavyFnMtDRT-L'
    destination = 'test.zip'
    g_download(file_id, destination)



from datetime import datetime
from typing import List, Optional
from dataclasses import dataclass, field

from learning_loop_node.detector.box_detection import BoxDetection
from learning_loop_node.detector.point_detection import PointDetection
from learning_loop_node.detector.segmentation_detection import SegmentationDetection


@dataclass
class Detections:
    box_detections: List[BoxDetection] = field(default_factory=list)
    point_detections: List[PointDetection] = field(default_factory=list)
    segmentation_detections: List[SegmentationDetection] = field(default_factory=list)
    tags: Optional[List[str]] = field(default_factory=list)
    date: Optional[str] = datetime.now().isoformat(sep='_', timespec='milliseconds')

    def __len__(self):
        return len(self.box_detections) + len(self.point_detections) + len(self.segmentation_detections)



from pydantic.main import BaseModel
from typing import List, Optional
from learning_loop_node.context import Context
from learning_loop_node.data_classes.category import Category
import os
import json


class ModelInformation(BaseModel):
    id: str
    host: Optional[str]
    organization: str
    project: str
    version: str
    categories: List[Category]
    resolution: Optional[int]
    model_root_path: Optional[str]

    @property
    def context(self):
        return Context(organization=self.organization, project=self.project)

    @staticmethod
    def load_from_disk(model_root_path: str):
        model_info_file_path = f'{model_root_path}/model.json'
        if not os.path.exists(model_info_file_path):
            raise FileExistsError(f"File '{model_info_file_path}' does not exist.")
        with open(model_info_file_path, 'r') as f:
            try:
                content = json.load(f)
            except:
                raise Exception(f"could not read model information from file '{model_info_file_path}'")
            try:
                model_information = ModelInformation.parse_obj(content)
                model_information.model_root_path = model_root_path
            except Exception as e:
                raise Exception(
                    f"could not parse model information from file '{model_info_file_path}'. \n {str(e)}")

        return model_information

    def save(self):
        with open(self.model_root_path + '/model.json', 'w') as f:
            f.write(self.json(exclude={'model_root_path'}))



from typing import Dict, Optional
from learning_loop_node.data_classes.category import Category
from learning_loop_node.trainer.hyperparameter import Hyperparameter
from pydantic import BaseModel
from typing import List


class TrainingData(BaseModel):
    image_data: List[dict] = []
    skipped_image_count: Optional[int] = 0
    categories: List[Category] = []
    hyperparameter: Optional[Hyperparameter]

    def image_ids(self):
        return [image['id'] for image in self.image_data]

    def train_image_count(self):
        return len([image for image in self.image_data if image['set'] == 'train'])

    def test_image_count(self):
        return len([image for image in self.image_data if image['set'] == 'test'])



from . import detections
from . import detections_upload_progress
from . import model_upload_progress
import json
from fastapi.encoders import jsonable_encoder
from learning_loop_node.trainer.training import Training
import os
from learning_loop_node.globals import GLOBALS

node_uuid = None


def init(uuid: str) -> None:
    global node_uuid
    node_uuid = uuid


def file_path() -> str:
    global node_uuid
    if not node_uuid:
        raise Exception('node_uuid not set. You have to call init(uuid: str) first')
    return f'{GLOBALS.data_folder}/last_training__{node_uuid}.json'


def save(training: Training):
    with open(file_path(), 'w') as f:
        json.dump(jsonable_encoder(training), f)


def load() -> Training:
    with open(file_path(), 'r') as f:
        return Training(**json.load(f))


def delete() -> None:
    if exists():
        os.remove(file_path())


def exists() -> bool:
    return os.path.exists(file_path())



from ..converter.converter import Converter
from ..status import State
from ..node import Node
from fastapi_utils.tasks import repeat_every
from icecream import ic
from ..loop import loop
import logging
from ..model_information import ModelInformation
from http import HTTPStatus
from fastapi.encoders import jsonable_encoder


class ConverterNode(Node):
    converter: Converter
    skip_check_state: bool = False
    bad_model_ids = []

    def __init__(self, name: str, converter: Converter, uuid: str = None):
        super().__init__(name, uuid)
        self.converter = converter

        @self.on_event("startup")
        @repeat_every(seconds=60, raise_exceptions=True, wait_first=False)
        async def check_state():
            if not self.skip_check_state:
                try:
                    await self.check_state()
                except:
                    logging.error('could not check state. Is loop reachable?')

    async def convert_model(self, model_information: ModelInformation):
        if model_information.id in self.bad_model_ids:
            logging.info(
                f'skipping bad model model {model_information.id} for {model_information.context.organization}/{model_information.context.project}.')
            return
        try:
            logging.info(
                f'converting model {jsonable_encoder(model_information)}')
            await self.converter.convert(model_information)
            logging.info('uploading model ')
            await self.converter.upload_model(model_information.context, model_information.id)
        except Exception as e:
            self.bad_model_ids.append(model_information.id)
            logging.error(
                f'could not convert model {model_information.id} for {model_information.context.organization}/{model_information.context.project}. Details: {str(e)}.')

    async def check_state(self):
        logging.debug(f'checking state: {self.status.state}')

        if self.status.state == State.Running:
            return
        self.status.state = State.Running
        try:
            await self.convert_models()
        except Exception as e:
            logging.error(str(e))

        self.status.state = State.Idle

    async def convert_models(self) -> None:
        async with loop.get('api/projects') as response:
            assert response.status == 200, f'Assert statuscode 200, but was {response.status}.'
            content = await response.json()
            projects = content['projects']

        for project in projects:
            organization_id = project['organization_id']
            project_id = project['project_id']

            async with loop.get(f'api{project["resource"]}') as response:
                if response.status != HTTPStatus.OK:
                    logging.error(
                        f'got bad response for {response.url}: {response.status}, {response.content}')
                    continue
                project_categories = (await response.json())['categories']

            path = f'api{project["resource"]}/models'
            async with loop.get(path) as models_response:
                assert models_response.status == 200
                content = await models_response.json()
                models = content['models']

                for model in models:
                    if (model['version']
                            and self.converter.source_format in model['formats']
                            and self.converter.target_format not in model['formats']
                            ):
                        # if self.converter.source_format in model['formats'] and project_id == 'drawingbot' and model['version'] == "6.0":
                        model_information = ModelInformation(
                            host=loop.base_url,
                            organization=organization_id,
                            project=project_id,
                            id=model['id'],
                            categories=project_categories,
                            version=model['version'],
                        )
                        await self.convert_model(model_information)

    async def send_status(self):
        # NOTE not yet implemented
        pass

    def get_state(self):
        return State.Idle



from dataclasses import dataclass
from typing import List


@dataclass
class Point:
    x: int
    y: int

    def __str__(self):
        return f'x:{int(self.x)} y: {int(self.y)}'


@dataclass
class Shape:
    points: List[Point]

    def __str__(self):
        return ', '. join([str(s) for s in self.points])


@dataclass
class SegmentationDetection:
    category_name: str
    shape: Shape
    model_name: str
    confidence: float
    category_id: str = ''

    @staticmethod
    def from_dict(detection: dict):
        category_id = detection['category_id'] if 'category_id' in detection else ''
        return SegmentationDetection(detection['category_name'], detection['shape'], detection['model_name'], detection['confidence'], category_id)

    def __str__(self):
        return f'shape:{str(self.shape)}, c: {self.confidence:.2f} -> {self.category_name}'



from enum import auto
import multiprocessing
from . import Detections
from . import Outbox
from .rest.operation_mode import OperationMode
from .detector import Detector
from .rest import detect, upload, operation_mode
from ..socket_response import SocketResponse
from ..inbox_filter.relevance_filter import RelevanceFilter
from ..rest import downloads
from ..node import Node
from ..status import State
from ..status import DetectionStatus, State
from ..context import Context
from ..globals import GLOBALS
from ..data_classes.category import Category
from ..model_information import ModelInformation
from fastapi.encoders import jsonable_encoder
from fastapi_utils.tasks import repeat_every
from typing import List, Union
import os
import contextlib
import logging
import subprocess
import asyncio
import numpy as np
from fastapi_socketio import SocketManager
from threading import Thread
from datetime import datetime
from icecream import ic
import shutil
from .. import environment_reader
from multiprocessing import Event


class DetectorNode(Node):
    update_frequency = 10

    def __init__(self, name: str, detector: Detector, uuid: str = None):
        super().__init__(name, uuid)
        self.detector = detector
        self.organization = environment_reader.organization()
        self.project = environment_reader.project()
        assert self.organization, 'Detector node needs an organization'
        assert self.project, 'Detector node needs an project'
        self.log.info(f'Using {self.organization}/{self.project}')
        self.operation_mode: OperationMode = OperationMode.Startup
        self.connected_clients: List[str] = []

        self.outbox: Outbox = Outbox()

        self.relevance_filter: RelevanceFilter = RelevanceFilter(self.outbox)
        self.target_model = None
        self.include_router(detect.router, tags=["detect"])
        self.include_router(upload.router, prefix="")
        self.include_router(operation_mode.router, tags=["operation_mode"])

        @self.on_event("startup")
        @repeat_every(seconds=self.update_frequency, raise_exceptions=False, wait_first=False)
        async def _check_for_update() -> None:
            try:
                await self.check_for_update()
            except:
                self.log.exception("error in '_check_for_update'")

        @self.on_event("startup")
        async def startup() -> None:
            try:
                self.log.info("received 'startup' event")
                self.outbox.start_continuous_upload()
                self._load_model()
            except:
                self.log.exception("error during 'startup'")

        sio = SocketManager(app=self)

        @self.sio.on("detect")
        async def _detect(sid, data) -> None:
            try:
                np_image = np.frombuffer(data['image'], np.uint8)
                return await self.get_detections(
                    raw_image=np_image,
                    camera_id=data.get('camera-id', None) or data.get('mac', None),
                    tags=data.get('tags', []),
                    autoupload=data.get('autoupload', None),
                )
            except Exception as e:
                self.log.exception('could not detect via socketio')
                with open('/tmp/bad_img_from_socket_io.jpg', 'wb') as f:
                    f.write(data['image'])
                return {'error': str(e)}

        @self.sio.on("info")
        async def _info(sid) -> None:
            if self.detector.current_model:
                return self.detector.current_model.__dict__
            return 'No model loaded'

        @self.sio.on('upload')
        async def _upload(sid, data):
            loop = asyncio.get_event_loop()
            try:
                await loop.run_in_executor(None, self.outbox.save, data['image'], Detections(), ['picked_by_system'])
            except Exception as e:
                self.log.exception('could not upload via socketio')
                return {'error': str(e)}

        @self.sio.event
        def connect(sid, environ, auth):
            self.connected_clients.append(sid)

        @self.on_event("shutdown")
        async def shutdown():
            await self.shutdown()

    async def shutdown(self):
        try:
            self.log.info("received 'shutdown' event")
            await self._disconnect_sio_clients()
            self.outbox.stop_continuous_upload()
        except:
            self.log.exception("error during 'shutdown'")

    def _load_model(self) -> None:
        try:
            self.detector.load_model()
        finally:
            self.operation_mode = OperationMode.Idle

    async def _disconnect_sio_clients(self):
        for sid in self.connected_clients:
            await self.sio.disconnect(sid)

    async def check_for_update(self):
        if self.operation_mode == OperationMode.Startup:
            return
        try:
            self.log.info(f'periodically checking operation mode. Current mode is {self.operation_mode}')
            update_to_model_id = await self.send_status()
            if self.detector.current_model:
                self.log.info(
                    f'Current model: {self.detector.current_model.version} with id {self.detector.current_model.id}')
            else:
                self.log.info(f'no model loaded')
            if self.operation_mode != OperationMode.Idle:
                self.log.info(f'not checking for updates; operation mode is {self.operation_mode}')
                return

            self.status.reset_error('update_model')
            if self.target_model is None:
                self.log.info(f'not checking for updates; no target model selected')
                return
            self.log.info('going to check for new updates')
            if not self.detector.current_model or self.target_model != self.detector.current_model.version:
                self.log.info(
                    f'Current model "{self.detector.current_model.version if self.detector.current_model else "-"}" needs to be updated to {self.target_model}')
                with pushd(GLOBALS.data_folder):
                    model_symlink = 'model'
                    target_model_folder = f'models/{self.target_model}'
                    shutil.rmtree(target_model_folder, ignore_errors=True)
                    os.makedirs(target_model_folder)

                    await downloads.download_model(
                        target_model_folder,
                        Context(organization=self.organization, project=self.project),
                        update_to_model_id,
                        self.detector.model_format
                    )
                    try:
                        os.unlink(model_symlink)
                        os.remove(model_symlink)
                    except:
                        pass
                    os.symlink(target_model_folder, model_symlink)
                    self.log.info(f'Updated symlink for model to {os.readlink(model_symlink)}')

                    await self.send_status()
                    self.reload(because='new model installed')
            else:
                self.log.info('Versions are identic. Nothing to do.')
        except Exception as e:
            self.log.exception(f'check_for_update failed')
            msg = e.cause if isinstance(e, downloads.DownloadError) else str(e)
            self.status.set_error('update_model', f'Could not update model: {msg}')
            await self.send_status()

    async def send_status(self) -> Union[str, bool]:
        if not self.sio_client.connected:
            self.log.error('could not send status -- we are not connected to the Learning Loop')
            return False
        status = DetectionStatus(
            id=self.uuid,
            name=self.name,
            state=self.status.state,
            errors=self.status._errors,
            uptime=int((datetime.now() - self.startup_time).total_seconds()),
            operation_mode=self.operation_mode,
            current_model=self.detector.current_model.version if self.detector.current_model else None,
            target_model=self.target_model,
            model_format=self.detector.model_format,
        )

        self.log.debug(f'sending status {status}')
        response = await self.sio_client.call('update_detector', (self.organization, self.project, jsonable_encoder(status)))
        socket_response = SocketResponse.from_dict(response)
        if not socket_response.success:
            self.log.error(f'Statusupdate failed: {response}')
            return False
        self.target_model = socket_response.payload['target_model_version']
        self.log.debug(f'After sending status. Target_model is {self.target_model}')
        return socket_response.payload['target_model_id']

    def get_state(self):
        return State.Online

    async def set_operation_mode(self, mode: OperationMode):
        self.operation_mode = mode
        await self.send_status()

    def reload(self, because: str):
        '''provide a cause for the reload'''

        print('########## reloading app because ' + because, flush=True)
        if os.path.isfile('/app/restart/restart.py'):
            subprocess.call(['touch', '/app/restart/restart.py'])
        else:
            subprocess.call(['touch', '/app/main.py'])

    async def get_detections(self, raw_image, camera_id: str, tags: str, autoupload: str = None):
        loop = asyncio.get_event_loop()
        detections = await loop.run_in_executor(None, self.detector.evaluate, raw_image)
        detections = self.add_category_id_to_detections(self.detector.model_info, detections)
        for detection in detections.segmentation_detections:
            detection.shape = ','.join([str(value) for p in detection.shape.points for _, value in p.__dict__.items()])
        info = "\n    ".join([str(d) for d in detections.box_detections +
                             detections.point_detections + detections.segmentation_detections])
        self.log.info(f'detected:\n    {info}')
        if camera_id is not None:
            tags.append(camera_id)
        if autoupload is None or autoupload == 'filtered':  # NOTE default is filtered
            Thread(target=self.relevance_filter.learn, args=(detections, camera_id, tags, raw_image)).start()
        elif autoupload == 'all':
            Thread(target=self.outbox.save, args=(raw_image, detections, tags)).start()
        elif autoupload == 'disabled':
            pass
        else:
            self.log.warning(f'unknown autoupload value {autoupload}')
        return jsonable_encoder(detections)

    async def upload_images(self, images: List[bytes]):
        loop = asyncio.get_event_loop()
        for image in images:
            await loop.run_in_executor(None, self.outbox.save, image, Detections(), ['picked_by_system'])

    def add_category_id_to_detections(self, model_info: ModelInformation, detections: Detections):
        def find_category_id_by_name(categories: List[Category], category_name: str):
            category_id = [category.id for category in categories if category.name == category_name]
            return category_id[0] if category_id else ''

        for box_detection in detections.box_detections:
            category_name = box_detection.category_name
            category_id = find_category_id_by_name(model_info.categories, category_name)
            box_detection.category_id = category_id
        for point_detection in detections.point_detections:
            category_name = point_detection.category_name
            category_id = find_category_id_by_name(model_info.categories, category_name)
            point_detection.category_id = category_id
        for segmentation_detection in detections.segmentation_detections:
            category_name = segmentation_detection.category_name
            category_id = find_category_id_by_name(model_info.categories, category_name)
            segmentation_detection.category_id = category_id
        return detections


@contextlib.contextmanager
def pushd(new_dir):
    previous_dir = os.getcwd()
    os.chdir(new_dir)
    try:
        yield
    finally:
        os.chdir(previous_dir)




from typing import List
import psutil
import os
import subprocess
import signal
from icecream import ic
import logging
import signal
from sys import platform


def create_signal_handler(sig=signal.SIGTERM):
    if platform == "linux" or platform == "linux2":
        # "The system will send a signal to the child once the parent exits for any reason (even sigkill)."
        # https://stackoverflow.com/a/19448096
        import ctypes
        libc = ctypes.CDLL("libc.so.6")

        def callable():
            os.setsid()
            return libc.prctl(1, sig)

        return callable
    return os.setsid


class Executor:

    def __init__(self, base_path) -> None:
        self.path = base_path
        os.makedirs(self.path, exist_ok=True)
        self.process = None

    def start(self, cmd: str):
        self.process = subprocess.Popen(
            f'cd {self.path}; {cmd} >> last_training.log 2>&1',
            shell=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            executable='/bin/bash',
            preexec_fn=create_signal_handler(),
        )

    def is_process_running(self):
        if self.process is None:
            return False

        if self.process.poll() is not None:
            return False

        try:
            psutil.Process(self.process.pid)
        except psutil.NoSuchProcess:
            return False

        return True

    def get_log(self) -> str:
        try:
            with open(f'{self.path}/last_training.log') as f:
                return f.read()
        except:
            return ''

    def get_log_by_lines(self) -> List[str]:
        try:
            with open(f'{self.path}/last_training.log') as f:
                return f.readlines()
        except:
            return []

    def stop(self):
        if self.process is None:
            logging.info('no process running ... nothing to stop')
            return

        logging.info('terminating process')

        try:
            os.killpg(os.getpgid(self.process.pid), signal.SIGTERM)
        except ProcessLookupError:
            pass

        self.process.terminate()
        out, err = self.process.communicate(timeout=3)

    @property
    def return_code(self):
        if not self.process:
            return None
        if self.is_process_running():
            return None
        return self.process.poll()



from learning_loop_node.trainer.tests.testing_trainer import TestingTrainer
import asyncio
from learning_loop_node.trainer.tests.states.state_helper import assert_training_state
from learning_loop_node.trainer.tests.states import state_helper
from learning_loop_node.trainer import active_training
from learning_loop_node.trainer.trainer import Trainer
from learning_loop_node.context import Context


error_key = 'upload_model'


def trainer_has_error(trainer: Trainer):
    return trainer.errors.has_error_for(error_key)


async def test_successful_upload(mocker):
    mock_upload_model_for_training(mocker, {'id': 'some_id'})

    state_helper.create_active_training_file()
    trainer = TestingTrainer()
    trainer.training = active_training.load()  # normally done by node

    train_task = asyncio.get_running_loop().create_task(trainer.upload_model())

    await assert_training_state(trainer.training, 'train_model_uploading', timeout=1, interval=0.001)
    await train_task

    assert trainer_has_error(trainer) == False
    assert trainer.training.training_state == 'train_model_uploaded'
    assert trainer.training.model_id_for_detecting == 'some_id'
    assert active_training.load() == trainer.training


async def test_abort_upload_model():
    state_helper.create_active_training_file(training_state='confusion_matrix_synced')
    trainer = TestingTrainer()
    trainer.training = active_training.load()  # normally done by node

    train_task = asyncio.get_running_loop().create_task(trainer.train(None, None))

    await assert_training_state(trainer.training, 'train_model_uploading', timeout=1, interval=0.001)

    trainer.stop()
    await asyncio.sleep(0.1)

    assert trainer.training == None
    assert active_training.exists() == False


async def test_bad_server_response_content(mocker):
    # without a running training the loop will not allow uploading.
    state_helper.create_active_training_file(training_state='confusion_matrix_synced')
    trainer = TestingTrainer()
    trainer.training = active_training.load()  # normally done by node

    train_task = asyncio.get_running_loop().create_task(trainer.train(None, None))

    await assert_training_state(trainer.training, 'train_model_uploading', timeout=1, interval=0.001)
    await assert_training_state(trainer.training, 'confusion_matrix_synced', timeout=2, interval=0.001)

    assert trainer_has_error(trainer)
    assert trainer.training.training_state == 'confusion_matrix_synced'
    assert trainer.training.model_id_for_detecting == None
    assert active_training.load() == trainer.training


async def test_mock_loop_response_example(mocker):
    mock_upload_model_for_training(mocker, 'some_data')

    state_helper.create_active_training_file()
    trainer = TestingTrainer()
    trainer.training = active_training.load()  # normally done by node

    result = await trainer._upload_model(Context(organization='zauberzeug', project='demo'))

    assert result == 'some_data'


def mock_upload_model_for_training(mocker, return_value):
    patched_call_return_value = asyncio.Future()
    patched_call_return_value.set_result(
        return_value)
    mocker.patch('learning_loop_node.rest.uploads.upload_model_for_training', return_value=patched_call_return_value)



from dataclasses import dataclass


@dataclass
class BoxDetection:
    category_name: str
    x: int
    y: int
    width: int
    height: int
    model_name: str
    confidence: float
    category_id: str = ''

    def __init__(self, category_name, x, y, width, height, net, confidence, category_id=''):
        self.category_id = category_id
        self.category_name = category_name
        self.x = x
        self.y = y
        self.width = width
        self.height = height
        self.model_name = net
        self.confidence = confidence

    def intersection_over_union(self, other_detection: 'BoxDetection') -> int:
        # https://www.pyimagesearch.com/2016/11/07/intersection-over-union-iou-for-object-detection/
        xA = max(self.x, other_detection.x)
        yA = max(self.y, other_detection.y)
        xB = min(self.x + self.width, other_detection.x + other_detection.width)
        yB = min(self.y + self.height, other_detection.y +
                 other_detection.height)

        interArea = max(xB - xA, 0) * max(yB - yA, 0)
        union = float(self._get_area() + other_detection._get_area() - interArea)
        if union == 0:
            print("WARNING: something went wrong while calculating iou")
            return 0

        return interArea / union

    def _get_area(self) -> int:
        return self.width * self.height

    @staticmethod
    def from_dict(detection: dict):
        category_id = detection['category_id'] if 'category_id' in detection else ''
        return BoxDetection(detection['category_name'], detection['x'], detection['y'], detection['width'], detection['height'], detection['model_name'], detection['confidence'], category_id)

    def __str__(self):
        return f'x:{int(self.x)} y: {int(self.y)}, w: {int(self.width)} h: {int(self.height)} c: {self.confidence:.2f} -> {self.category_name}'



import pytest
from learning_loop_node import DetectorNode
import requests
from learning_loop_node.detector.tests.conftest import get_outbox_files


def test_detector_path(test_detector_node: DetectorNode):
    assert test_detector_node.outbox.path.startswith('/tmp')


@pytest.mark.parametrize('test_detector_node', [True], indirect=True)
async def test_sio_detect(test_detector_node, sio_client):
    with open('detector/tests/test.jpg', 'rb') as f:
        image_bytes = f.read()
    result = await sio_client.call('detect', {'image': image_bytes})
    assert len(result['box_detections']) == 1
    assert result['box_detections'][0]['category_name'] == 'some_category_name'
    assert result['box_detections'][0]['category_id'] == 'some_id'

    assert len(result['point_detections']) == 1
    assert result['point_detections'][0]['category_name'] == 'some_category_name'
    assert result['point_detections'][0]['category_id'] == 'some_id'

    assert len(result['segmentation_detections']) == 1
    assert result['segmentation_detections'][0]['category_name'] == 'some_category_name'
    assert result['segmentation_detections'][0]['category_id'] == 'some_id'


@pytest.mark.parametrize('grouping_key', ['mac', 'camera_id'])
@pytest.mark.parametrize('test_detector_node', [True], indirect=True)
def test_rest_detect(test_detector_node: DetectorNode, grouping_key: str):
    image = {('file', open('detector/tests/test.jpg', 'rb'))}
    headers = {grouping_key: '0:0:0:0', 'tags':  'some_tag'}
    response = requests.post(f'http://localhost:{pytest.detector_port}/detect', files=image, headers=headers)
    assert response.status_code == 200
    result = response.json()
    assert len(result['box_detections']) == 1
    assert result['box_detections'][0]['category_name'] == 'some_category_name'
    assert result['box_detections'][0]['category_id'] == 'some_id'

    assert len(result['point_detections']) == 1
    assert result['point_detections'][0]['category_name'] == 'some_category_name'
    assert result['point_detections'][0]['category_id'] == 'some_id'

    assert len(result['segmentation_detections']) == 1
    assert result['segmentation_detections'][0]['category_name'] == 'some_category_name'
    assert result['segmentation_detections'][0]['category_id'] == 'some_id'


def test_rest_upload(test_detector_node: DetectorNode):
    assert len(get_outbox_files(test_detector_node.outbox)) == 0

    image = {('files', open('detector/tests/test.jpg', 'rb'))}
    response = requests.post(f'http://localhost:{pytest.detector_port}/upload', files=image)
    assert response.status_code == 200
    assert len(get_outbox_files(test_detector_node.outbox)) == 2, 'There should be one image and one .json file.'


@pytest.mark.parametrize('test_detector_node', [True], indirect=True)
async def test_sio_upload(test_detector_node: DetectorNode, sio_client):
    assert len(get_outbox_files(test_detector_node.outbox)) == 0

    with open('detector/tests/test.jpg', 'rb') as f:
        image_bytes = f.read()
    result = await sio_client.call('upload', {'image': image_bytes})
    assert result == None
    assert len(get_outbox_files(test_detector_node.outbox)) == 2, 'There should be one image and one .json file.'



from dataclasses import dataclass
from typing import List, Optional
from pydantic import BaseModel
from enum import Enum
from learning_loop_node.context import Context
from learning_loop_node.data_classes.category import Category


class Point(BaseModel):
    x: int
    y: int


class Shape(BaseModel):
    points: List[Point]


class SegmentationAnnotation(BaseModel):
    id: str
    shape: Shape
    image_id: str
    category_id: str


class EventType(str, Enum):
    LeftMouseDown = 'left_mouse_down',
    RightMouseDown = 'right_mouse_down',
    MouseMove = 'mouse_move',
    LeftMouseUp = 'left_mouse_up',
    RightMouseUp = 'right_mouse_up',
    KeyUp = 'key_up',
    KeyDown = 'key_down',


class AnnotationData(BaseModel):
    coordinate: Point
    event_type: EventType
    context: Context
    image_uuid: str
    category: Category
    is_shift_key_pressed: Optional[bool]
    key_up: Optional[str] = None
    key_down: Optional[str] = None
    # keyboard_modifiers: Optional[List[str]]
    # new_annotation_uuid: Optional[str]
    # edit_annotation_uuid: Optional[str]


class UserInput(BaseModel):
    frontend_id: str
    data: AnnotationData


class ToolOutput(BaseModel):
    svg: str
    annotation: Optional[SegmentationAnnotation]



from typing import List, Optional

from ..detector.outbox.outbox import Outbox
from ..detector.detections import Detections
from .relevance_group import RelevanceGroup


class RelevanceFilter():

    def __init__(self, outbox: Outbox) -> None:
        self.groups: dict[str, RelevanceGroup] = {}
        self.outbox: Outbox = outbox

    def learn(
        self,
        detections: Detections,
        camera_id: str,
        tags: Optional[str],
        raw_image: bytes
    ) -> List[str]:
        {group.forget_old_detections() for (_, group) in self.groups.items()}
        if camera_id not in self.groups:
            self.groups[camera_id] = RelevanceGroup()
        causes = self.groups[camera_id].add_detections(detections)
        if len(detections) >= 80:
            causes.append('unexpectedObservationsCount')
        if len(causes) > 0:
            tags.extend(causes)
            self.outbox.save(raw_image, detections, tags)
        return causes

    def reset(self) -> None:
        self.learners = {}



from typing import Union

from learning_loop_node.detector.segmentation_detection import SegmentationDetection
from ..detector.box_detection import BoxDetection
from ..detector.point_detection import PointDetection
from datetime import datetime, timedelta


class Observation():

    def __init__(self, detection: Union[BoxDetection, PointDetection, SegmentationDetection]):
        self.detection = detection
        self.last_seen = datetime.now()

    def update_last_seen(self):
        self.last_seen = datetime.now()

    def is_older_than(self, forget_time_in_seconds):
        return self.last_seen < datetime.now() - timedelta(seconds=forget_time_in_seconds)



import multiprocessing
import pytest
from learning_loop_node import DetectorNode, ModelInformation
from learning_loop_node.detector import Outbox
from learning_loop_node.data_classes import Category
from testing_detector import TestingDetector
import uvicorn
from multiprocessing import Process, log_to_stderr
import logging
import icecream
import os
import socket
import asyncio
from typing import Generator
import socketio
from glob import glob

logging.basicConfig(level=logging.INFO)

# show ouptut from uvicorn server https://stackoverflow.com/a/66132186/364388
log_to_stderr(logging.INFO)

icecream.install()


def pytest_configure():
    pytest.detector_port = 5000


def should_have_segmentations(request) -> bool:
    should_have_segmentations = False
    try:
        should_have_segmentations = request.param
    except:
        pass
    return should_have_segmentations


@pytest.fixture()
async def test_detector_node(request):
    os.environ['ORGANIZATION'] = 'zauberzeug'
    os.environ['PROJECT'] = 'demo'

    model_info = ModelInformation(id='some_uuid', host='some_host', organization='zauberzeug',
                                  project='test', version='1', categories=[Category(id='some_id', name='some_category_name')])
    segmentations = should_have_segmentations(request)

    det = TestingDetector(segmentation_detections=segmentations)
    det.init(model_info=model_info, model_root_path='')
    node = DetectorNode(name='test', detector=det)
    await port_is(free=True)

    multiprocessing.set_start_method('fork', force=True)
    assert multiprocessing.get_start_method() == 'fork'
    proc = Process(target=uvicorn.run,
                   args=(node,),
                   kwargs={
                       "host": "127.0.0.1",
                       "port": pytest.detector_port,

                   }, daemon=True)
    proc.start()
    await port_is(free=False)
    yield node

    try:
        await node.shutdown()
    except:
        logging.exception('error while shutting down node')

    try:
        proc.kill()
    except:  # for python 3.6
        proc.terminate()
    proc.join()

# from https://stackoverflow.com/a/52872579/364388


def is_port_in_use(port):
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        return s.connect_ex(('localhost', port)) == 0


async def port_is(free: bool):
    for i in range(10):
        if not free and is_port_in_use(pytest.detector_port):
            return
        if free and not is_port_in_use(pytest.detector_port):
            return
        else:
            await asyncio.sleep(0.5)
    raise Exception(f'port {pytest.detector_port} is {"not" if free else ""} free')


@pytest.fixture()
async def sio_client() -> Generator:
    sio = socketio.AsyncClient()
    try_connect = True
    retry_count = 0
    while try_connect:
        try:
            await sio.connect(f"ws://localhost:{pytest.detector_port}", socketio_path="/ws/socket.io")
            try_connect = False
        except:
            logging.warning('trying again')
            await asyncio.sleep(1)
        retry_count += 1
        if retry_count > 10:
            raise Exception('Max Retry')

    yield sio
    await sio.disconnect()


def get_outbox_files(outbox: Outbox):
    files = glob(f'{outbox.path}/**/*', recursive=True)
    return [file for file in files if os.path.isfile(file)]



from typing import Optional, Union, Any
import logging
from learning_loop_node.model_information import ModelInformation
from learning_loop_node.detector.detections import Detections
from learning_loop_node.globals import GLOBALS


class Detector():

    def __init__(self, model_format: str) -> None:
        self.model_format: str = model_format
        self.current_model: Optional[ModelInformation] = None
        self.target_model: Optional[str] = None

    def load_model(self):
        try:
            model_information = ModelInformation.load_from_disk(f'{GLOBALS.data_folder}/model')
            try:
                self.init(model_information)
                self.current_model = model_information
                logging.info(f'Successfully loaded model {self.current_model}')
            except Exception:
                logging.error(f'Could not init model {model_information}')
                raise
        except Exception:
            self.current_model = None
            logging.exception('An error occured during loading model.')

    def init(self,  model_info: ModelInformation):
        raise NotImplementedError()

    def evaluate(self, image: Any) -> Detections:
        raise NotImplementedError()



from typing import List, Optional
from pydantic import BaseModel
from enum import Enum
from dataclasses import dataclass


class State(str, Enum):
    Idle = "idle"
    Offline = "offline"
    Online = "online"
    Preparing = "preparing"
    Running = "running"
    Stopping = "stopping"
    Detecting = 'detecting'
    Uploading = 'uploading'


class Status(BaseModel):
    id: str
    name: str
    state: Optional[State] = State.Offline
    uptime: Optional[int] = 0
    _errors: Optional[dict] = {}

    def set_error(self, key: str, value: str):
        self._errors[key] = value

    def reset_error(self, key: str):
        try:
            del self._errors[key]
        except AttributeError:
            pass
        except KeyError:
            pass

    def reset_all_errors(self):
        for key in list(self._errors.keys()):
            self.reset_error(key)


class AnnotationNodeStatus(Status):
    capabilities: List[str]


@dataclass
class DetectionStatus():
    id: str
    name: str
    state: Optional[State]
    errors: Optional[dict]
    uptime: Optional[int]

    model_format: str
    current_model: Optional[str]
    target_model: Optional[str]
    operation_mode: Optional[str]



from learning_loop_node.trainer.tests.testing_trainer import TestingTrainer
import asyncio
from learning_loop_node.trainer.tests.states.state_helper import assert_training_state
from learning_loop_node.trainer.tests.states import state_helper
from learning_loop_node.trainer import active_training
from learning_loop_node.tests.test_helper import condition
from asyncio import coroutine


async def test_successful_training():
    state_helper.create_active_training_file(training_state='train_model_downloaded')
    trainer = TestingTrainer()
    trainer.training = active_training.load()  # normally done by node

    train_task = asyncio.get_running_loop().create_task(trainer.train(None, None))

    await assert_training_state(trainer.training, 'training_running', timeout=1, interval=0.001)
    assert trainer.train_task.__name__ == 'start_training'

    trainer.executor.stop()  # NOTE normally a training terminates itself e.g
    await assert_training_state(trainer.training, 'training_finished', timeout=1, interval=0.001)

    assert trainer.training.training_state == 'training_finished'
    assert active_training.load() == trainer.training


async def test_stop_running_training():
    state_helper.create_active_training_file(training_state='train_model_downloaded')
    trainer = TestingTrainer()
    trainer.training = active_training.load()  # normally done by node

    train_task = asyncio.get_running_loop().create_task(trainer.train(None, None))

    await condition(lambda: trainer.executor and trainer.executor.is_process_running(), timeout=1, interval=0.01)
    await assert_training_state(trainer.training, 'training_running', timeout=1, interval=0.001)
    assert trainer.train_task.__name__ == 'start_training'

    trainer.stop()
    await assert_training_state(trainer.training, 'training_finished', timeout=1, interval=0.001)

    assert trainer.training.training_state == 'training_finished'
    assert active_training.load() == trainer.training


async def test_training_can_maybe_resumed():
    # NOTE e.g. when a node-computer is restarted
    state_helper.create_active_training_file(training_state='train_model_downloaded')
    trainer = TestingTrainer(can_resume=True)
    trainer.training = active_training.load()  # normally done by node

    train_task = asyncio.get_running_loop().create_task(trainer.train(None, None))

    await condition(lambda: trainer.executor and trainer.executor.is_process_running(), timeout=1, interval=0.01)
    await assert_training_state(trainer.training, 'training_running', timeout=1, interval=0.001)
    assert trainer.train_task.__name__ == 'resume'

    trainer.executor.stop()  # NOTE normally a training terminates itself e.g
    await assert_training_state(trainer.training, 'training_finished', timeout=1, interval=0.001)

    assert trainer.training.training_state == 'training_finished'
    assert active_training.load() == trainer.training



from typing import List, Optional, Tuple
from learning_loop_node.context import Context
from learning_loop_node.rest.downloader import DataDownloader
from learning_loop_node.task_logger import create_task
from learning_loop_node.trainer.training_data import TrainingData
import logging
import os
from icecream import ic


class TrainingsDownloader():
    downloader: DataDownloader
    data_query_params: str

    def __init__(self, context: Context, data_query_params: Optional[str] = 'state=complete'):
        self.data_query_params = data_query_params
        self.downloader = DataDownloader(context)

    async def download_training_data(self, image_folder: str) -> Tuple[List[dict], int]:
        image_ids = await self.downloader.fetch_image_ids(query_params=self.data_query_params)
        image_data, skipped_image_count = await self.download_images_and_annotations(image_ids, image_folder)
        return (image_data, skipped_image_count)

    async def download_images_and_annotations(self, image_ids: List[str], image_folder: str) -> TrainingData:
        await self.downloader.download_images(image_ids, image_folder)
        image_data = await self.downloader.download_images_data(image_ids)
        logging.info('filtering corrupt images')  # download only safes valid images
        valid_image_data: List[dict] = []
        skipped_image_count = 0
        for i in image_data:
            if os.path.isfile(f'{image_folder}/{i["id"]}.jpg'):
                valid_image_data.append(i)
            else:
                skipped_image_count += 1
        logging.info(f'Done downloading image data for {len(image_data)} images.')
        return (valid_image_data, skipped_image_count)




from learning_loop_node.context import Context
from learning_loop_node.node import Node
from learning_loop_node.loop import loop
from fastapi import APIRouter,  Request, HTTPException
from fastapi.responses import PlainTextResponse
import logging

from learning_loop_node.trainer.trainer import Trainer

router = APIRouter()


@router.post("/controls/detect/{organization}/{project}/{version}")
async def operation_mode(organization: str, project: str, version: str, request: Request):
    '''
    Example Usage
        curl -X POST localhost/controls/detect/<organization>/<project>/<model_version>
    '''
    path = f'/api/{organization}/projects/{project}/models'
    async with loop.get(path) as response:
        if response.status != 200:
            raise HTTPException(404, 'could not load latest model')
        models = (await response.json())['models']
        model_id = next(m for m in models if m['version'] == version)['id']
        logging.info(model_id)
    trainer: Trainer = request.app.trainer
    await trainer.do_detections(Context(organization=organization, project=project), model_id)
    return "OK"



from dataclasses import dataclass
from typing import Optional
from pydantic import BaseModel


class BasicModel(BaseModel):
    confusion_matrix: Optional[dict]
    meta_information: Optional[dict]


class Model(BaseModel):
    id: str
    confusion_matrix: Optional[dict]
    parent_id: Optional[str]
    train_image_count: Optional[int]
    test_image_count: Optional[int]
    trainer_id: Optional[str]
    hyperparameters: Optional[str]


@dataclass
class PretrainedModel():
    name: str
    label: str
    description: str



import os
import shutil
from typing import List, Optional
from pydantic.main import BaseModel
from learning_loop_node.node import Node
from learning_loop_node.model_information import ModelInformation
from learning_loop_node.rest import downloads, uploads
from icecream import ic
import json


class Converter(BaseModel):
    model_folder: Optional[str]
    source_format: str
    target_format: str

    async def convert(self, model_information: ModelInformation) -> None:
        project_folder = Node.create_project_folder(model_information.context)

        self.model_folder = Converter.create_model_folder(project_folder, model_information.id)
        await downloads.download_model(self.model_folder, model_information.context, model_information.id, self.source_format)

        with open(f'{self.model_folder}/model.json', 'r') as f:
            content = json.load(f)
            if 'resolution' in content:
                model_information.resolution = content['resolution']

        await self._convert(model_information)

    async def _convert(self, model_information: ModelInformation) -> None:
        raise NotImplementedError()

    def get_converted_files(self, model_id) -> List[str]:
        raise NotImplementedError()

    async def upload_model(self, context, model_id: str) -> bool:
        files = self.get_converted_files(model_id)
        await uploads.upload_model(context, files, model_id, self.target_format)

    @staticmethod
    def create_convert_folder(project_folder: str) -> str:
        image_folder = f'{project_folder}/images'
        os.makedirs(image_folder, exist_ok=True)
        return image_folder

    @staticmethod
    def create_model_folder(project_folder: str, model_id: str) -> str:
        model_folder = f'{project_folder}/{model_id}'
        shutil.rmtree(model_folder, ignore_errors=True)  # cleanup
        os.makedirs(model_folder, exist_ok=True)
        return model_folder




from learning_loop_node.trainer.tests.testing_trainer import TestingTrainer
from learning_loop_node.trainer.trainer import Trainer
import asyncio
from learning_loop_node.trainer.tests.states.state_helper import assert_training_state
from learning_loop_node.trainer.tests.states import state_helper
from learning_loop_node.trainer import active_training
import os
import logging


async def test_downloading_is_successfull():
    state_helper.create_active_training_file(training_state='data_downloaded')
    trainer = Trainer(model_format='mocked')
    trainer.training = active_training.load()  # normally done by node

    download_task = asyncio.get_running_loop().create_task(trainer.download_model())
    await assert_training_state(trainer.training, 'train_model_downloading', timeout=1, interval=0.001)
    await assert_training_state(trainer.training, 'train_model_downloaded', timeout=1, interval=0.001)

    assert trainer.training.training_state == 'train_model_downloaded'
    assert active_training.load() == trainer.training

    # file on disk
    assert os.path.exists(f'{trainer.training.training_folder}/base_model.json')
    assert os.path.exists(f'{trainer.training.training_folder}/file_1.txt')
    assert os.path.exists(f'{trainer.training.training_folder}/file_2.txt')


async def test_abort_download_model():
    state_helper.create_active_training_file(training_state='data_downloaded')
    trainer = TestingTrainer()
    trainer.training = active_training.load()  # normally done by node

    train_task = asyncio.get_running_loop().create_task(trainer.train(None, None))
    await assert_training_state(trainer.training, 'train_model_downloading', timeout=1, interval=0.001)

    trainer.stop()
    await asyncio.sleep(0.1)

    assert trainer.training == None
    assert active_training.exists() == False


async def test_downloading_failed():
    state_helper.create_active_training_file(training_state='data_downloaded',
                                             base_model_id='00000000-0000-0000-0000-000000000000')  # bad model id)
    trainer = Trainer(model_format='mocked')
    trainer.training = active_training.load()  # normally done by node

    train_task = asyncio.get_running_loop().create_task(trainer.train(None, None))
    await assert_training_state(trainer.training, 'train_model_downloading', timeout=1, interval=0.001)
    await assert_training_state(trainer.training, 'data_downloaded', timeout=1, interval=0.001)

    assert trainer.errors.has_error_for('download_model')
    assert trainer.training is not None
    assert trainer.training.training_state == 'data_downloaded'
    assert active_training.load() == trainer.training



import pytest
import os
from fastapi.encoders import jsonable_encoder
from learning_loop_node.context import Context
from learning_loop_node.annotation_node.data_classes import AnnotationData, EventType, Point, UserInput, EventType
from learning_loop_node.annotation_node.annotation_node import AnnotationNode
from learning_loop_node.annotation_node.annotation_tool import EmptyAnnotationTool
from learning_loop_node.data_classes.category import Category, CategoryType
from icecream import ic


def default_user_input() -> UserInput:
    annotation_data = AnnotationData(
        coordinate=Point(x=0, y=0),
        event_type=EventType.LeftMouseDown,
        context=Context(organization='zauberzeug', project='pytest'),
        image_uuid='285a92db-bc64-240d-50c2-3212d3973566',
        category=Category(id='some_id', name='category_1', description='',
                          hotkey='', color='', type=CategoryType.Segmentation)
    )
    return UserInput(frontend_id='some_id', data=annotation_data)


async def test_image_download(create_project):
    image_path = '/tmp/learning_loop_lib_data/zauberzeug/pytest/images/285a92db-bc64-240d-50c2-3212d3973566.jpg'

    assert os.path.exists(image_path) == False

    node = AnnotationNode(name="", uuid="", tool=EmptyAnnotationTool())
    input = default_user_input()
    _ = await node.handle_user_input(jsonable_encoder(input))

    assert os.path.exists(image_path) == True



from typing import Any, Optional
from pydantic import BaseModel
import asyncio
import functools
import logging
import traceback


class SocketResponse(BaseModel):
    success: bool
    error_msg: Optional[str]
    payload: Optional[Any]

    @staticmethod
    def for_failure(error_msg: str):
        return SocketResponse(success=False, error_msg=error_msg)

    @staticmethod
    def for_success(payload: Optional[str] = ''):
        return SocketResponse(success=True, payload=payload)

    @staticmethod
    def from_bool(value: bool):
        return SocketResponse(success=value)

    @staticmethod
    def from_dict(value: dict):
        try:
            return SocketResponse.parse_obj(value)
        except:
            logging.exception(f'Error parsing SocketResponse: value : {value}')
            raise


def ensure_socket_response(func):
    @functools.wraps(func)
    async def wrapper_ensure_socket_response(*args, **kwargs):
        try:
            if asyncio.iscoroutinefunction(func):
                value = await func(*args, **kwargs)
            else:
                value = func(*args, **kwargs)

            if isinstance(value, str):
                return SocketResponse.for_success(value).__dict__
            elif isinstance(value, bool):
                return SocketResponse.from_bool(value).__dict__
            elif isinstance(value, SocketResponse):
                return value
            elif (args[0] in ['connect', 'disconnect', 'connect_error']):
                return value
            else:
                raise Exception(f"Returntype for sio must be str or bool or SocketResponse', but was {type(value)}'")
        except Exception:
            error = traceback.print_exc()
            trace = ''.join(traceback.format_stack())
            logging.error(
                f'\nAn error occured for {args[0]}:  \
                \nStacktrace: \
                \n{trace} \
                \nError: \
                \n {str(error)} \n'
            )

            return SocketResponse.for_failure(str(error)).__dict__

    return wrapper_ensure_socket_response



from typing import List
from learning_loop_node.trainer.training import Training
import os


def file_path(training: Training) -> str:
    return f'{training.training_folder}/model_uploading_progress.txt'


def save(training: Training, formats: List[str]) -> None:
    with open(file_path(training), 'w') as f:
        f.write(','.join(formats))


def load(training: Training) -> List[str]:
    if not exists(training):
        return []
    with open(file_path(training), 'r') as f:
        return f.read().split(',')


def delete(training) -> None:
    path = file_path(training)
    if exists(training):
        os.remove(file_path(training))


def exists(training) -> None:
    return os.path.exists(file_path(training))



from datetime import datetime, timedelta
from learning_loop_node.inbox_filter.observation import Observation
from learning_loop_node.detector.box_detection import BoxDetection
import time


def test_aging():
    _detection = Observation(BoxDetection(None, None, None, None, None, None, None))
    time.sleep(0.2)
    assert _detection.is_older_than(0.1) == True
    assert _detection.is_older_than(0.3) == False


def test_calculate_iou():
    one = BoxDetection('dirt', 10, 0, 30, 100, None, 30)
    two = BoxDetection('dirt', 20, 0, 30, 100, None, 61)
    three = BoxDetection('dirt', 0, 30, 10, 10, None, 61)

    assert one.intersection_over_union(two) == 0.5
    assert one.intersection_over_union(three) == 0.0


def test_update_last_seen():
    observation = Observation(BoxDetection(None, None, None, None, None, None, None))
    observation.last_seen = datetime.now() - timedelta(seconds=.5)
    assert observation.is_older_than(0.5) == True



from typing import List
import pytest

import logging
from learning_loop_node.model_information import ModelInformation
from learning_loop_node.converter.converter import Converter
from learning_loop_node.converter.converter_node import ConverterNode
from learning_loop_node.tests import test_helper


class MockedConverter(Converter):
    models: List[ModelInformation] = []

    async def _convert(self, model_information: ModelInformation) -> None:
        self.models.append(model_information)


@pytest.fixture()
def create_project():
    test_helper.LiveServerSession().delete(f"/api/zauberzeug/projects/pytest?keep_images=true")
    project_configuration = {'project_name': 'pytest', 'box_categories': 1,  'point_categories': 1, 'inbox': 0, 'annotate': 0, 'review': 0, 'complete': 0, 'image_style': 'plain',
                             'thumbs': False, 'trainings': 1}
    assert test_helper.LiveServerSession().post(f"/api/zauberzeug/projects/generator",
                                                json=project_configuration).status_code == 200
    yield
    test_helper.LiveServerSession().delete(f"/api/zauberzeug/projects/pytest?keep_images=true")


async def test_meta_information(create_project):
    model_id = await test_helper.get_latest_model_id()

    converter = MockedConverter(source_format='mocked', target_format='test')
    node = ConverterNode(name='test', converter=converter)
    await node.convert_models()

    pytest_project_model = [m for m in converter.models if m.id == model_id][0]

    categories = pytest_project_model.categories
    assert len(categories) == 2
    assert categories[0].type == 'box'
    assert categories[1].type == 'point'



from enum import Enum
from learning_loop_node.trainer.training_data import TrainingData
from pydantic import BaseModel
from typing import Optional
from learning_loop_node.context import Context
import json
from fastapi.encoders import jsonable_encoder


class Training(BaseModel):
    base_model_id: Optional[str]
    id: str
    context: Context

    project_folder: str
    images_folder: str

    training_folder: Optional[str]

    data: Optional[TrainingData]

    training_number: Optional[int]

    training_state: Optional[str]

    model_id_for_detecting: Optional[str]


class TrainingOut(BaseModel):
    confusion_matrix: Optional[dict]
    train_image_count: Optional[int]
    test_image_count: Optional[int]
    trainer_id: Optional[str]
    hyperparameters: Optional[dict]


class State(str, Enum):
    Initialized = 'initialized'
    DataDownloading = 'data_downloading'
    DataDownloaded = 'data_downloaded'
    TrainModelDownloading = 'train_model_downloading'
    TrainModelDownloaded = 'train_model_downloaded'
    TrainingRunning = 'training_running'
    TrainingFinished = 'training_finished'
    ConfusionMatrixSyncing = 'confusion_matrix_syncing'
    ConfusionMatrixSynced = 'confusion_matrix_synced'
    TrainModelUploading = 'train_model_uploading'
    TrainModelUploaded = 'train_model_uploaded'
    Detecting = 'detecting'
    Detected = 'detected'
    DetectionUploading = 'detection_uploading'
    ReadyForCleanup = 'ready_for_cleanup'



from dataclasses import dataclass
import numpy as np


@dataclass
class PointDetection:
    category_name: str
    x: int
    y: int
    model_name: str
    confidence: float
    category_id: str = ''

    def __init__(self, category_name, x, y, net, confidence, category_id=''):
        self.category_name = category_name
        self.x = x
        self.y = y
        self.model_name = net
        self.confidence = confidence
        self.category_id = category_id

    @staticmethod
    def from_dict(detection: dict):
        category_id = detection['category_id'] if 'category_id' in detection else ''
        return PointDetection(detection['category_name'], detection['x'], detection['y'], detection['model_name'], detection['confidence'], category_id)


    def distance(self, other: 'PointDetection') -> float:
        return np.sqrt((other.x - self.x)**2 + (other.y - self.y)**2)

    def __str__(self):
        return f'x:{int(self.x)} y: {int(self.y)}, c: {self.confidence:.2f} -> {self.category_name}'



from fastapi import APIRouter, Request, File, UploadFile
from typing import List


router = APIRouter()


@router.post("/upload")
async def upload_image(request: Request, files: List[UploadFile] = File(...)):
    """
    Example Usage

        curl -X POST -F 'files=@test.jpg' "http://localhost:/upload"
    """
    raw_files = [await file.read() for file in files]
    await request.app.upload_images(raw_files)
    return 200, "OK"



from learning_loop_node.trainer.tests.testing_trainer import TestingTrainer
from learning_loop_node.trainer.trainer import Trainer
import asyncio
from learning_loop_node.trainer.tests.states.state_helper import assert_training_state
from learning_loop_node.trainer.tests.states import state_helper
from learning_loop_node.trainer import active_training
from learning_loop_node.context import Context

error_key = 'prepare'


def trainer_has_error(trainer: Trainer):
    return trainer.errors.has_error_for(error_key)


async def test_preparing_is_successful():
    state_helper.create_active_training_file()
    trainer = Trainer(model_format='mocked')
    trainer.training = active_training.load()  # normally done by node

    await trainer.prepare()

    assert trainer_has_error(trainer) == False
    assert trainer.training.training_state == 'data_downloaded'
    assert trainer.training.data is not None
    assert active_training.load() == trainer.training


async def test_abort_preparing():
    state_helper.create_active_training_file()
    trainer = TestingTrainer()
    trainer.training = active_training.load()  # normally done by node

    train_task = asyncio.get_running_loop().create_task(trainer.train(None, None))
    await assert_training_state(trainer.training, 'data_downloading', timeout=1, interval=0.001)

    trainer.stop()
    await asyncio.sleep(0.1)

    assert trainer.training == None
    assert active_training.exists() == False


async def test_request_error():
    state_helper.create_active_training_file(context=Context(
        organization='zauberzeug', project='some_bad_project'))
    trainer = Trainer(model_format='mocked')
    trainer.training = active_training.load()  # normally done by node

    train_task = asyncio.get_running_loop().create_task(trainer.train(None, None))
    await assert_training_state(trainer.training, 'data_downloading', timeout=3, interval=0.001)
    await assert_training_state(trainer.training, 'initialized', timeout=3, interval=0.001)

    assert trainer_has_error(trainer)
    assert trainer.training is not None
    assert trainer.training.training_state == 'initialized'
    assert active_training.load() == trainer.training



from typing import Optional
from pydantic.main import BaseModel


class ErrorConfiguration(BaseModel):
    begin_training: Optional[bool] = False
    save_model: Optional[bool] = False
    get_new_model: Optional[bool] = False
    crash_training: Optional[bool] = False



from learning_loop_node.trainer import active_training
from learning_loop_node.trainer.tests.testing_trainer import TestingTrainer
import asyncio
from learning_loop_node.trainer.tests.states.state_helper import assert_training_state
from learning_loop_node.trainer.tests.states import state_helper
from learning_loop_node.trainer import active_training
from learning_loop_node.trainer.trainer import Trainer


error_key = 'detecting'


def trainer_has_error(trainer: Trainer):
    return trainer.errors.has_error_for(error_key)


async def test_successfull_detecting():
    state_helper.create_active_training_file(training_state='train_model_uploaded',
                                             model_id_for_detecting='7f5eabb4-227a-e7c7-8f0b-f825cc47340d')  # version 1.2 of demo project
    trainer = TestingTrainer()
    trainer.training = active_training.load()  # normally done by node

    detect_task = asyncio.get_running_loop().create_task(trainer.do_detections())

    await assert_training_state(trainer.training, 'detecting', timeout=1, interval=0.001)
    await assert_training_state(trainer.training, 'detected', timeout=2, interval=0.001)

    assert trainer_has_error(trainer) == False
    assert trainer.training.training_state == 'detected'
    assert active_training.load() == trainer.training
    assert active_training.detections.exists(trainer.training)


async def test_detecting_can_be_aborted():
    state_helper.create_active_training_file(training_state='train_model_uploaded',
                                             model_id_for_detecting='7f5eabb4-227a-e7c7-8f0b-f825cc47340d')  # version 1.2 of demo project
    trainer = TestingTrainer()
    trainer.training = active_training.load()  # normally done by node
    training = trainer.training

    train_task = asyncio.get_running_loop().create_task(trainer.train(None, None))

    await assert_training_state(trainer.training, 'detecting', timeout=1, interval=0.001)
    trainer.stop()
    await asyncio.sleep(0.1)

    assert trainer.training is None
    assert active_training.detections.exists(training) is False
    assert active_training.exists() == False


async def test_model_not_downloadable_error():
    state_helper.create_active_training_file(training_state='train_model_uploaded',
                                             model_id_for_detecting='00000000-0000-0000-0000-000000000000')  # bad model id
    trainer = TestingTrainer()
    trainer.training = active_training.load()  # normally done by node

    train_task = asyncio.get_running_loop().create_task(trainer.train(None, None))

    await assert_training_state(trainer.training, 'detecting', timeout=1, interval=0.001)
    await assert_training_state(trainer.training, 'train_model_uploaded', timeout=1, interval=0.001)

    assert trainer_has_error(trainer)
    assert trainer.training.training_state == 'train_model_uploaded'
    assert trainer.training.model_id_for_detecting == '00000000-0000-0000-0000-000000000000'
    assert active_training.load() == trainer.training


def test_save_load_detections():
    detections = [{'some_key': 'some_value'}]
    state_helper.create_active_training_file()
    trainer = TestingTrainer()
    trainer.training = active_training.load()

    active_training.detections.save(trainer.training, detections)
    assert active_training.detections.exists(trainer.training)

    stored_detections = active_training.detections.load(trainer.training)
    assert stored_detections == detections

    active_training.detections.delete(trainer.training)
    assert active_training.detections.exists(trainer.training) is False



from abc import abstractmethod
from pydantic import BaseModel
from learning_loop_node.annotation_node.data_classes import UserInput, ToolOutput
import logging


class AnnotationTool(BaseModel):

    @abstractmethod
    async def handle_user_input(self, user_input: UserInput, history: dict) -> ToolOutput:
        pass

    @abstractmethod
    def create_empty_history(self) -> dict:
        pass

    @abstractmethod
    def logout_user(self, sid: str):
        pass


class EmptyAnnotationTool():
    async def handle_user_input(self, user_input: UserInput, history: dict) -> ToolOutput:
        return ToolOutput(svg="", shape={})

    def create_empty_history(self) -> dict:
        return {}

    def logout_user(self, sid: str) -> bool:
        return True



import numpy as np
from learning_loop_node import DetectorNode
import pytest
from learning_loop_node.detector.tests.conftest import get_outbox_files
import asyncio


@pytest.mark.parametrize('autoupload,expected_file_count', [(None, 2), ('all', 4)])
async def test_filter_is_used_by_node(test_detector_node: DetectorNode, autoupload, expected_file_count):
    assert test_detector_node.outbox.path.startswith('/tmp')
    assert len(get_outbox_files(test_detector_node.outbox)) == 0

    image = np.fromfile('detector/tests/test.jpg', np.uint8)
    _ = await test_detector_node.get_detections(image, '00:.....', tags=[], autoupload=autoupload)
    # NOTE adding second images with identical detections
    _ = await test_detector_node.get_detections(image, '00:.....', tags=[], autoupload=autoupload)
    await asyncio.sleep(1)  # files are stored asynchronously

    assert len(get_outbox_files(test_detector_node.outbox)) == expected_file_count,\
        'There should be 1 image and 1 .json file for every detection in the outbox'



from typing import List, Union
import aiohttp
from learning_loop_node.context import Context
from learning_loop_node.loop import loop
import logging
from icecream import ic


async def upload_model(context: Context, files: List[str], model_id: str, format: str) -> None:
    data = aiohttp.FormData()

    for file_name in files:
        data.add_field('files',  open(file_name, 'rb'))
    async with loop.put(f'api/{context.organization}/projects/{context.project}/models/{model_id}/{format}/file', data=data) as response:
        if response.status != 200:
            msg = f'---- could not upload model with id {model_id} and format {format}. Details: {await response.text()}'
            raise Exception(msg)
        else:
            logging.info(f'---- uploaded model with id {model_id} and format {format}.')


async def upload_model_for_training(context: Context, files: List[str], training_number: int, format: str) -> Union[dict, None]:
    data = aiohttp.FormData()

    for file_name in files:
        data.add_field('files',  open(file_name, 'rb'))
    async with loop.put(f'api/{context.organization}/projects/{context.project}/trainings/{training_number}/models/latest/{format}/file', data=data) as response:
        if response.status != 200:
            msg = f'---- could not upload model for training {training_number} and format {format}. Details: {await response.text()}'
            logging.error(msg)
            response.raise_for_status()
        else:
            uploaded_model = await response.json()
            logging.info(
                f'---- uploaded model for training {training_number} and format {format}. Model id is {uploaded_model}')
            return uploaded_model



import os


def organization(default: str = None):
    return os.environ.get('LOOP_ORGANIZATION', default) or os.environ.get('ORGANIZATION', default)


def project(default: str = None):
    return os.environ.get('LOOP_PROJECT', default) or os.environ.get('PROJECT', default)




from learning_loop_node.trainer.training import TrainingOut
from learning_loop_node.socket_response import SocketResponse
from fastapi.encoders import jsonable_encoder
import logging
import socketio
import asyncio


async def try_sync_model(trainer: any, trainer_node_uuid: str, sio_client: socketio.AsyncClient):
    try:
        model = trainer.get_new_model()
    except Exception as e:
        logging.exception('error while getting new model')
        raise Exception(f'Could not get new model : {str(e)}')
    logging.debug(f'new model {model}')

    if model:
        response = await sync_model(trainer, trainer_node_uuid, sio_client, model)

        if not response.success:
            error_msg = f'Error for update_training: Response from loop was : {response.__dict__}'
            logging.error(error_msg)
            raise Exception(error_msg)


async def sync_model(trainer, trainer_node_uuid, sio_client, model):
    current_training = trainer.training
    new_training = TrainingOut(
        trainer_id=trainer_node_uuid,
        confusion_matrix=model.confusion_matrix,
        train_image_count=current_training.data.train_image_count(),
        test_image_count=current_training.data.test_image_count(),
        hyperparameters=trainer.hyperparameters)

    await asyncio.sleep(0.1)  # NOTE needed for tests.

    result = await sio_client.call('update_training', (current_training.context.organization, current_training.context.project, jsonable_encoder(new_training)))
    response = SocketResponse.from_dict(result)

    if response.success:
        logging.info(f'successfully updated training {jsonable_encoder(new_training)}')
        trainer.on_model_published(model)
    return response



"""original copied from https://quantlane.com/blog/ensure-asyncio-task-exceptions-get-logged/"""

from typing import Any, Awaitable, Optional, TypeVar, Tuple

import asyncio
import functools
import logging


T = TypeVar('T')


def create_task(
    coroutine: Awaitable[T],
    *,
    loop: Optional[asyncio.AbstractEventLoop] = None,
) -> 'asyncio.Task[T]':  # This type annotation has to be quoted for Python < 3.9, see https://www.python.org/dev/peps/pep-0585/
    '''
    This helper function wraps a ``loop.create_task(coroutine())`` call and ensures there is
    an exception handler added to the resulting task. If the task raises an exception it is logged
    using the provided ``logger``, with additional context provided by ``message`` and optionally
    ``message_args``.
    '''

    logger = logging.getLogger(__name__)
    message = 'Task raised an exception'
    message_args = ()
    if loop is None:
        loop = asyncio.get_running_loop()
    task = loop.create_task(coroutine)
    task.add_done_callback(
        functools.partial(_handle_task_result, logger=logger, message=message, message_args=message_args)
    )
    return task


def _handle_task_result(
    task: asyncio.Task,
    *,
    logger: logging.Logger,
    message: str,
    message_args: Tuple[Any, ...] = (),
) -> None:
    try:
        task.result()
    except asyncio.CancelledError:
        pass  # Task cancellation should not be logged as an error.
    # Ad the pylint ignore: we want to handle all exceptions here so that the result of the task
    # is properly logged. There is no point re-raising the exception in this callback.
    except Exception:  # pylint: disable=broad-except
        logger.exception(message, *message_args)



import logging
from learning_loop_node.tests import test_helper
from learning_loop_node.loop import loop
import pytest
import shutil
import icecream
from learning_loop_node.globals import GLOBALS
import os
icecream.install()
logging.basicConfig(level=logging.DEBUG)


@pytest.fixture()
def create_project():
    test_helper.LiveServerSession().delete(f"/api/zauberzeug/projects/pytest?keep_images=true")
    project_configuration = {'project_name': 'pytest', 'inbox': 0, 'annotate': 0, 'review': 0, 'complete': 3, 'image_style': 'beautiful',
                             'box_categories': 2, 'point_categories': 2, 'segmentation_categories': 2, 'thumbs': False, 'tags': 0, 'trainings': 1, 'box_detections': 3, 'box_annotations': 0}
    assert test_helper.LiveServerSession().post(f"/api/zauberzeug/projects/generator",
                                                json=project_configuration).status_code == 200
    yield
    test_helper.LiveServerSession().delete(f"/api/zauberzeug/projects/pytest?keep_images=true")


@pytest.fixture(autouse=True, scope='session')
def clea_loggers():
    yield
    # see https://github.com/pytest-dev/pytest/issues/5502
    """Remove handlers from all loggers"""
    import logging
    loggers = [logging.getLogger()] + list(logging.Logger.manager.loggerDict.values())
    for logger in loggers:
        handlers = getattr(logger, 'handlers', [])
        for handler in handlers:
            logger.removeHandler(handler)


@pytest.fixture(autouse=True, scope='function')
def loop_session():
    loop.session = None
    yield
    loop.session = None


@pytest.fixture(autouse=True, scope='function')
def data_folder():
    GLOBALS.data_folder = '/tmp/learning_loop_lib_data'
    shutil.rmtree(GLOBALS.data_folder, ignore_errors=True)
    os.makedirs(GLOBALS.data_folder, exist_ok=True)
    yield
    shutil.rmtree(GLOBALS.data_folder, ignore_errors=True)
